# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from towhee.array import Array
from towhee.dataframe.dataframe_v2 import DataFrame


class TestDataframe(unittest.TestCase):
    """
    Test dataframe basic function
    """

    def test_constructors(self):

        def get_columns():
            return ['digit', 'letter']

        def get_tuples():
            return [(0, 'a'), (1, 'b'), (2, 'c')]

        def get_arrays():
            return [Array([0, 1, 2]), Array(['a', 'b', 'c'])]

        def get_dict():
            return {'digit': Array([0, 1, 2]), 'letter': Array(['a', 'b', 'c'])}

        def check_data(df):
            for i in range(3):
                self.assertEqual(df['digit'][i], i)
                self.assertEqual(df['letter'][i], chr(ord('a') + i))
                self.assertEqual(df[i][0], i)
                self.assertEqual(df[i][1], chr(ord('a') + i))
            for i, row in enumerate(df.iter()):
                self.assertEqual(row[0], i)
                self.assertEqual(row[1], chr(ord('a') + i))

        # empty df
        df = DataFrame('my_df')
        df.seal()
        self.assertEqual(df.name, 'my_df')

        # from list[tuple]
        data = get_tuples()
        columns = get_columns()
        df = DataFrame('my_df', data, columns)
        df.seal()
        check_data(df)

        # from list[towhee.Array]
        data = get_arrays()
        columns = get_columns()
        df = DataFrame('my_df', data, columns)
        df.seal()
        check_data(df)

        # from dict[str, towhee.Array]
        data = get_dict()
        df = DataFrame('my_df', data)
        df.seal()
        check_data(df)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple
import torch
import numpy

from towhee.operator import Operator


class PytorchTransformerVecOperator(Operator):
    """
    Pytorch transformer vector operator, the output is embedding instead of prediction
        Args:
            model(nn.Module):
                Model object
    """

    def __init__(self, model) -> None:
        super().__init__()
        self._model = model

    def __call__(self, img_tensor: torch.Tensor) -> NamedTuple('Outputs', [('embedding', numpy.ndarray)]):
        self._model.eval()
        with torch.no_grad():
            outputs = self._model(img_tensor).squeeze(0)
        Outputs = NamedTuple('Outputs', [('embedding', numpy.ndarray)])
        return Outputs(outputs)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from pathlib import Path

from towhee.tests.mock_operators import ADD_OPERATOR_PATH, load_local_operator
from towhee.operator import SharedType, PyOperator, NNOperator, Operator

_MOCK_OPERATOR_DIR = os.path.join(Path(__file__).parent.parent, 'mock_operators')
NN_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, 'nn_operator')
PY_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, 'py_operator')


class TestOperator(unittest.TestCase):
    """
    Simple operator test
    """
    def test_nnoperator(self):
        self.assertTrue(issubclass(NNOperator, Operator))

        nn_operator = load_local_operator('nn_operator', NN_OPERATOR_PATH)
        pt_op = nn_operator.TestNNOperator()
        tf_op = nn_operator.TestNNOperator('tensorflow')
        self.assertIsInstance(pt_op, NNOperator)
        self.assertIsInstance(tf_op, NNOperator)
        self.assertEqual(pt_op.framework, 'pytorch')
        self.assertEqual(tf_op.framework, 'tensorflow')

        pt_op.framework = 'tensorflow'
        self.assertEqual(pt_op.framework, tf_op.framework)

    def test_pyoperator(self):
        self.assertTrue(issubclass(PyOperator, Operator))

        py_operator = load_local_operator('py_operator', PY_OPERATOR_PATH)
        op = py_operator.TestPyOperator()

        self.assertIsInstance(op, PyOperator)

    def test_add_function(self):
        add_operator = load_local_operator('add_operator', ADD_OPERATOR_PATH)
        op = add_operator.AddOperator(1)
        self.assertTrue(isinstance(op, PyOperator))
        self.assertEqual(op(1).sum, 2)
        self.assertEqual(op.shared_type, SharedType.Shareable)

        op2 = add_operator.AddOperator(3)
        self.assertEqual(op2(1).sum, 4)
        self.assertEqual(op2.shared_type, SharedType.Shareable)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class EmbeddingOutput:
    """
    Container for embedding extractor.
    """
    def __init__(self):
        self.embeddings = []

    def __call__(self, module, module_in, module_out):
        self.embeddings.append(module_out)

    def clear(self):
        self.embeddings = []


class EmbeddingExtractor:
    """
    Embedding extractor from a layer
    Args:
        model (`nn.Module`):
            Model used for inference.
    """
    def __init__(self, model):
        self.modules = model.modules()
        self.modules_list = list(model.named_modules(remove_duplicate=False))
        self.modules_dict = dict(model.named_modules(remove_duplicate=False))
        self.emb_out = EmbeddingOutput()

    def disp_modules(self):
        """
        Display the the modules of the model.
        """
        for idx, m in enumerate(self.modules):
            if idx:
                print(idx, '->', m)

    def register(self, layer_name: str):
        """
        Registration for embedding extraction.
        Args:
            layer_name (`str`):
                Name of the layer from which the embedding is extracted.
        """
        if layer_name in self.modules_dict:
            layer_indx = 0
            for layer in self.modules:
                layer_indx = layer_indx + 1
                if self.modules_list[layer_indx][0] == layer_name:
                    _ = layer.register_forward_hook(self.emb_out)
                    break
        else:
            raise ValueError('layer_name not in modules')



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import torchvision

from pathlib import Path
from PIL import Image
from towhee.models.layers.netvlad import NetVLAD, EmbedNet
from torchvision.models import resnet18
from torch import nn


cache_path = Path(__file__).parent.parent.parent.resolve()
test_image = cache_path.joinpath('data/dataset/kaggle_dataset_small/train/001cdf01b096e06d78e9e5112d419397.jpg')


class NetvladTest(unittest.TestCase):
    def setUp(self) -> None:
        # Discard layers at the end of base network
        encoder = resnet18(pretrained=True)
        base_model = nn.Sequential(
            encoder.conv1,
            encoder.bn1,
            encoder.relu,
            encoder.maxpool,
            encoder.layer1,
            encoder.layer2,
            encoder.layer3,
            encoder.layer4,
        )
        self.dim = list(base_model.parameters())[-1].shape[0]  # last channels (512)
        self.num_clusters = 32

        # Define model for embedding, K clusters determined by num_clusters
        net_vlad = NetVLAD(num_clusters=self.num_clusters, dim=self.dim, alpha=1.0)
        self.model = EmbedNet(base_model, net_vlad)
        img_pil = Image.open(test_image)
        self.img_tensor = torchvision.transforms.ToTensor()(img_pil).unsqueeze(0)

    def test_netvlad(self):
        # x = torch.rand(1, 3, 256, 436)
        output = self.model(self.img_tensor)
        self.assertEqual(self.num_clusters * self.dim, output.shape[1])


if __name__ == '__main__':
    unittest.main(verbosity=1)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Pytorch version of 'An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale'.
Jax version: https://github.com/google-research/vision_transformer.
"""


def get_base_config():
    """
    Base ViT config ViT
    """
    return dict(
      dim=768,
      ff_dim=3072,
      num_heads=12,
      num_layers=12,
      attention_dropout_rate=0.0,
      dropout_rate=0.1,
      representation_size=768,
      classifier='token'
    )


def get_b16_config():
    """
    Returns the ViT-B/16 configuration.
    """
    config = get_base_config()
    config.update(dict(patches=(16, 16)))
    return config


def get_b32_config():
    """Returns the ViT-B/32 configuration."""
    config = get_b16_config()
    config.update(dict(patches=(32, 32)))
    return config


def get_l16_config():
    """Returns the ViT-L/16 configuration."""
    config = get_base_config()
    config.update(dict(
        patches=(16, 16),
        dim=1024,
        ff_dim=4096,
        num_heads=16,
        num_layers=24,
        attention_dropout_rate=0.0,
        dropout_rate=0.1,
        representation_size=1024
    ))
    return config


def get_l32_config():
    """
    Returns the ViT-L/32 configuration.
    """
    config = get_l16_config()
    config.update(dict(patches=(32, 32)))
    return config


def drop_head_variant(config):
    config.update(dict(representation_size=None))
    return config



# Copyright 2021 Ross Wightman and Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
from itertools import repeat
import collections.abc


# From PyTorch internals
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse


to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple






# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List
import time
import random
import weakref
import threading

from towhee.engine.thread_pool_task_executor import ThreadPoolTaskExecutor


class TaskScheduler(ABC):
    """Task scheduler abstract interface.

    Args:
        task_execs: (`List[towhee.TaskExecutor]`)
            A list of task executors that the `Engine` manages. Also should be
            continuously changing as new executors are acquired.
    """

    def __init__(self, task_execs: List[ThreadPoolTaskExecutor]):
        self._task_execs = task_execs
        self._graph_ctx_refs = []
        self._lock = threading.Lock()
        self._need_stop = False

    @abstractmethod
    def stop(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def join(self) -> None:
        raise NotImplementedError

    def schedule_forever(self, sleep_ms: int = 1000):
        """Runs the a single schedule step in a loop.

        sleep_ms: (`int`)
            Milliseconds to sleep after completing a single scheduling step.
        """
        while not self._need_stop:
            self.schedule_step()
            # TODO(fzliu): compute runtime for bottleneck operator
            time.sleep(sleep_ms / 1000)

    @abstractmethod
    def schedule_step(self):
        raise NotImplementedError


class BasicScheduler(TaskScheduler):
    """
    Basic scheduler.

    This scheduler will start all ops, and has no scheduling logic, so the
    schedule_step does nothing.
    """

    def register(self, graph_ctx):
        for op in graph_ctx.op_ctxs.values():
            executor = self._find_optimal_exec()
            op.start(executor)

        with self._lock:
            self._graph_ctx_refs.append(weakref.ref(graph_ctx))

    def stop(self):
        for g_ctx_ref in self._graph_ctx_refs:
            g_ctx = g_ctx_ref()
            if g_ctx is not None:
                g_ctx.stop()

    def join(self):
        for g_ctx_ref in self._graph_ctx_refs:
            g_ctx = g_ctx_ref()
            if g_ctx is not None:
                g_ctx.join()

    def schedule_step(self):
        """
        Do nothing
        """
        pass

    def _find_optimal_exec(self):
        """
        Acquires the least busy instance of `TaskExecutor` that can still execute the operator.
        """
        return random.choice(self._task_execs)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import torch

from towhee.models.layers.spatial_temporal_cls_positional_encoding import SpatialTemporalClsPositionalEncoding


class SpaTempoClsPosEncodeTest(unittest.TestCase):
    def test_spatial_temporal_cls_positional_encoding(self):
        # Test with cls token.
        batch_dim = 4
        dim = 16
        video_shape = (1, 2, 4)
        video_sum = video_shape[0] * video_shape[1] * video_shape[2]
        has_cls = True
        model = SpatialTemporalClsPositionalEncoding(
            embed_dim=dim,
            patch_embed_shape=video_shape,
            has_cls=has_cls,
        )
        fake_input = torch.rand(batch_dim, video_sum, dim)
        output = model(fake_input)
        output_gt_shape = (batch_dim, video_sum + 1, dim)
        self.assertTrue(tuple(output.shape) == output_gt_shape)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import torch
import torchvision

from towhee.models.layers.sam import SAM


class TestOperator(unittest.TestCase):
    """
    A test for SAM.
    """

    def setUp(self):
        self.model = torchvision.models.resnet50(pretrained=True)
        self.base_optimizer = torch.optim.SGD

    def test_sam(self):
        optimizer = SAM(self.model.parameters(), self.base_optimizer, lr=0.1, momentum=0.9)
        self.assertEqual(8, len(optimizer.base_optimizer.param_groups[0]))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import NamedTuple
from towhee.operator import Operator, SharedType


class AddOperatorV2(Operator):
    """
    Add operator that can take in undefined amount of inputs and saves results to '<input_name>_sum'
    column of result dataframe.
    """

    def __init__(self, factor: int) -> None:
        super().__init__()
        self._factor = factor

    def __call__(self, **kwargs):
        cols = []
        ret = []
        for key, value in kwargs.items():
            cols.append((key + '_sum', type(value)))
            ret.append(value + self._factor)

        outputs = NamedTuple('Outputs', cols)
        return outputs(*ret) #pylint: disable=not-callable

    @property
    def shared_type(self):
        return SharedType.Shareable



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from towhee.engine.operator_runner.runner_base import RunnerBase, RunnerStatus


class MockRunner(RunnerBase):
    def process_step(self) -> bool:
        self._set_finished()
        return True


class TestRunnerBase(unittest.TestCase):
    """
    Runner base test
    """

    def test_runner_base(self):
        mock_runner = MockRunner('mock', 0, 'add_operator',
                                 'mock_operators', {'num': 1})
        self.assertEqual(mock_runner.status, RunnerStatus.IDLE)
        self.assertTrue(mock_runner.is_idle())
        mock_runner.process()
        self.assertEqual(mock_runner.status, RunnerStatus.FINISHED)
        self.assertEqual(mock_runner.op_name, 'add_operator')
        self.assertEqual(mock_runner.hub_op_id, 'mock_operators')
        self.assertEqual(mock_runner.op_args, {'num': 1})



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import requests
import os

from towhee.data.video_info import get_video_info


class VideoInfoTest(unittest.TestCase):
    url = 'https://dl.fbaipublicfiles.com/pytorchvideo/projects/theatre.webm'
    filename = url.split('/', maxsplit=1)[-1]
    root = os.path.dirname(os.path.abspath(__file__))
    file_path = root + '/' + filename
    r = requests.get(url)
    with open(file_path, 'wb') as f:
        f.write(r.content)
        f.close()

    def test_video_info(self):
        info = get_video_info(self.file_path)
        self.assertEqual(info.fps, 30)
        self.assertEqual(info.width, 1280)
        self.assertEqual(info.height, 720)
        self.assertEqual(info.frame_count, 7128)
        self.assertEqual(info.duration, 237.6)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import unittest

from towhee.models.layers.sequence_pool import SequencePool


class SequencePoolTest(unittest.TestCase):
    def test_sequence_pool(self):
        model = SequencePool('cls')
        fake_input = torch.rand(8, 10, 10)
        output = model(fake_input)
        self.assertTrue(torch.equal(output, fake_input[:, 0]))
        model = SequencePool('mean')
        output = model(fake_input)
        self.assertTrue(torch.equal(output, fake_input.mean(1)))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import torch

from towhee.models.layers.multi_scale_transformer_block import MultiScaleBlock


class MultiScaleTransformerBlockTest(unittest.TestCase):
    def test_multi_scale_transformer_block(self):
        # Change of output dimension.
        block = MultiScaleBlock(10, 20, 2)
        seq_len = 21
        c_dim = 10
        batch_dim = 8
        fake_input = torch.rand(batch_dim, seq_len, c_dim)
        input_shape = (2, 2, 5)
        output, output_shape = block(fake_input, input_shape)
        gt_shape_tensor = torch.rand(8, seq_len, 20)
        self.assertTrue(output.shape == gt_shape_tensor.shape)
        self.assertTrue(output_shape == input_shape)

        # Test pooling.
        block = MultiScaleBlock(10, 20, 2, stride_q=(2, 2, 1))
        c_dim = 10
        batch_dim = 8
        fake_input = torch.rand(batch_dim, seq_len, c_dim)
        input_shape = [2, 2, 5]
        output, output_shape = block(fake_input, input_shape)
        gt_shape_tensor = torch.rand(8, int((seq_len - 1) / 2 / 2) + 1, 20)
        gt_out_shape = [1, 1, 5]
        self.assertTrue(output.shape == gt_shape_tensor.shape)
        self.assertTrue(output_shape == gt_out_shape)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import numpy
import torch

from typing import NamedTuple
from pathlib import Path

import torchvision

from towhee.operator import Operator

cache_path = Path(__file__).parent.parent.parent.resolve()


def decode_predictions(preds):
    fpath = cache_path.joinpath('data/dataset/imagenet_index_path.json')
    with open(fpath, encoding='ascii') as f:
        class_index = json.load(f)

    results = tuple(class_index[str(preds)])[1]
    return results


class PyTorchImageClassificationOperator(Operator):
    """
    PyTorch model operator base
    """

    def __init__(self, model_name) -> None:
        super().__init__()
        model_func = getattr(torchvision.models, model_name)
        self._model = model_func(pretrained=True)
        self._model.eval()

    def __call__(self, img_tensor: torch.Tensor) -> NamedTuple('Outputs', [('embedding', numpy.ndarray),
                                                                           ('breed', str)]):
        Outputs = NamedTuple('Outputs', [('embedding', numpy.ndarray), ('breed', str)])
        outputs = self._model(img_tensor)
        _, preds = torch.max(outputs, 1)
        results = decode_predictions(int(preds))
        return Outputs(outputs.detach().numpy(), results)

    def train(self):
        """
        For training model
        """
        pass



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import NamedTuple, List
from towhee.operator import Operator, SharedType


class SumOperator(Operator):
    """
    Flat input list.
    """

    def __init__(self):
        pass

    def __call__(self, inputs: List[NamedTuple('input', [('num', int)])]):
        Outputs = NamedTuple('Outputs', [('sum', int)])
        nums = [item.num for item in inputs]
        output = Outputs(sum(nums))
        return [output, output]

    @property
    def shared_type(self):
        return SharedType.Shareable



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import torch

from towhee.models.layers.gatedmlp import GatedMlp

class GatedMlpTest(unittest.TestCase):
    def test_gatedmlp(self):
        in_features = 32
        hidden_features = 64
        out_features = 128

        mod = GatedMlp(in_features, hidden_features, out_features)

        fake_input = torch.rand(1, in_features)
        output = mod(fake_input)

        gt_output_shape = torch.Size([1, out_features])
        self.assertTrue(output.shape == gt_output_shape)




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class _ArrayRef:
    """
    `_ArrayRef` maintains `DataFrame` to `Array` references.
    """

    def __init__(self):
        self._read_ref_offsets = {}
        self._ref_id_allocator = 0

    def add_reader(self) -> int:
        """
        Add a read reference to this `Array`

        Returns:
            (`int`)
                reference id
        """
        ref_id = self._ref_id_allocator
        self._read_ref_offsets[ref_id] = 0
        self._ref_id_allocator += 1
        return ref_id

    def remove_reader(self, ref_id: int):
        """
        Remove a read reference from this `Array`

        Args:
            ref_id (`int`):
                The reference ID
        """
        self._read_ref_offsets.pop(ref_id, None)

    def update_reader_offset(self, ref_id: int, offset: int):
        if ref_id in self._read_ref_offsets:
            self._read_ref_offsets[ref_id] = offset
        else:
            raise KeyError('reference id %d not fount' % (ref_id))

    @property
    def min_reader_offsets(self):
        if self._read_ref_offsets:
            return min(self._read_ref_offsets.values())
        else:
            return 0



# coding : UTF-8
from towhee import pipeline

class TestPipelineInvalid:
    """ Test case of invalid pipeline interface """

    def test_pipeline_no_params(self):
        """
        target: test pipeline for invalid scenario
        method:  call pipeline with no params
        expected: raise exception
        """
        try:
            embedding_pipeline = pipeline()
        except TypeError as e:
            assert "pipeline() missing 1 required positional argument: 'task'" in str(e)

    def test_pipeline_wrong_params(self):
        """
        target: test pipeline for invalid scenario
        method:  call pipeline with wrong pipeline name
        expected: raise exception
        """
        wrong_pipeline = "wrong-embedding"
        try:
            embedding_pipeline = pipeline(wrong_pipeline)
        except Exception as e:
            assert "Incorrect pipeline format" in str(e)

class TestPipelineValid:
    """ Test case of valid pipeline interface """

    def test_pipeline(self, pipeline_name):
        """
        target: test pipeline for image normal case
        method:  call pipeline with right pipeline name
        expected: return object
        """
        embedding_pipeline = pipeline(pipeline_name)
        assert "_pipeline" in dir(embedding_pipeline)








# Copyright 2021 Ross Wightman and Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
import math
from typing import Callable, Tuple
from functools import partial

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from towhee.models.utils.general_utils import to_2tuple
from towhee.models.layers.conv2d_same import conv2d_same
from towhee.models.layers.padding_functions import get_padding_value

def get_condconv_initializer(initializer: Callable, num_experts: int, expert_shape: Tuple[int, int, int, int]):
    def condconv_initializer(weight):
        """CondConv initializer function."""
        num_params = np.prod(expert_shape)
        if (len(weight.shape) != 2 or weight.shape[0] != num_experts or
                weight.shape[1] != num_params):
            raise (ValueError(
                'CondConv variables must have shape [num_experts, num_params]'))
        for i in range(num_experts):
            initializer(weight[i].view(expert_shape))
    return condconv_initializer

class CondConv2d(nn.Module):
    """
    Conditionally Parameterized Convolution

    Inspired by: https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/condconv/condconv_layers.py
    Grouped convolution hackery for parallel execution of the per-sample kernel filters inspired by this discussion:
    https://github.com/pytorch/pytorch/issues/17983

    Args:
        in_channels (`int`):
            Number of channels in the input image
        out_channels (`int`):
            Number of channels produced by the convolution.
        kernel_size (`int`):
            Size of the convolving kernel.
        stride (`int`):
            Stride of the convolution.
        padding (`str`):
            Padding added to all four sides of the input.
        dilation (`int`):
            Spacing between kernel elements.
        groups (`int`):
            Number of blocked connections from input channels to output channels.
        bias (`bool`):
            If True, adds a learnable bias to the output.
        num_experts (`int`):
            The number of expert kernels and biases.
    """
    __constants__ = ['in_channels', 'out_channels', 'dynamic_padding']

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int=3,
                 stride: int=1, padding: str='', dilation: int=1, groups: int=1, bias: bool=False, num_experts: int=4) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = to_2tuple(kernel_size)
        self.stride = to_2tuple(stride)
        padding_val, is_padding_dynamic = get_padding_value(
            padding, kernel_size, stride=stride, dilation=dilation)
        self.dynamic_padding = is_padding_dynamic  # if in forward to work with torchscript
        self.padding = to_2tuple(padding_val)
        self.dilation = to_2tuple(dilation)
        self.groups = groups
        self.num_experts = num_experts

        self.weight_shape = (self.out_channels, self.in_channels // self.groups) + self.kernel_size
        weight_num_param = 1
        for wd in self.weight_shape:
            weight_num_param *= wd
        self.weight = torch.nn.Parameter(torch.Tensor(self.num_experts, weight_num_param))

        if bias:
            self.bias_shape = (self.out_channels,)
            self.bias = torch.nn.Parameter(torch.Tensor(self.num_experts, self.out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        init_weight = get_condconv_initializer(
            partial(nn.init.kaiming_uniform_, a=math.sqrt(5)), self.num_experts, self.weight_shape)
        init_weight(self.weight)
        if self.bias is not None:
            fan_in = np.prod(self.weight_shape[1:])
            bound = 1 / math.sqrt(fan_in)
            init_bias = get_condconv_initializer(
                partial(nn.init.uniform_, a=-bound, b=bound), self.num_experts, self.bias_shape)
            init_bias(self.bias)

    def forward(self, x: torch.Tensor, routing_weights: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape
        weight = torch.matmul(routing_weights, self.weight)
        new_weight_shape = (b * self.out_channels, self.in_channels // self.groups) + self.kernel_size
        weight = weight.view(new_weight_shape)
        bias = None
        if self.bias is not None:
            bias = torch.matmul(routing_weights, self.bias)
            bias = bias.view(b * self.out_channels)
        # move batch elements with channels so each batch element can be efficiently convolved with separate kernel
        x = x.view(1, b * c, h, w)
        if self.dynamic_padding:
            out = conv2d_same(
                x, weight, bias, stride=self.stride, padding=self.padding,
                dilation=self.dilation, groups=self.groups * b)
        else:
            out = F.conv2d(
                x, weight, bias, stride=self.stride, padding=self.padding,
                dilation=self.dilation, groups=self.groups * b)
        out = out.permute([1, 0, 2, 3]).view(b, self.out_channels, out.shape[-2], out.shape[-1])
        return out




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy
import torch
import torchvision

from typing import NamedTuple, List
from pathlib import Path
from towhee.operator import Operator

cache_path = Path(__file__).parent.parent.parent.resolve()

COCO_INSTANCE_CATEGORY_NAMES = [
    '__background__', 'person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus',
    'train', 'truck', 'boat', 'traffic light', 'fire hydrant', 'N/A', 'stop sign',
    'parking meter', 'bench', 'bird', 'cat', 'dog', 'horse', 'sheep', 'cow',
    'elephant', 'bear', 'zebra', 'giraffe', 'N/A', 'backpack', 'umbrella', 'N/A', 'N/A',
    'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball',
    'kite', 'baseball bat', 'baseball glove', 'skateboard', 'surfboard', 'tennis racket',
    'bottle', 'N/A', 'wine glass', 'cup', 'fork', 'knife', 'spoon', 'bowl',
    'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza',
    'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'N/A', 'dining table',
    'N/A', 'N/A', 'toilet', 'N/A', 'tv', 'laptop', 'mouse', 'remote', 'keyboard', 'cell phone',
    'microwave', 'oven', 'toaster', 'sink', 'refrigerator', 'N/A', 'book',
    'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush'
]
lb_names = COCO_INSTANCE_CATEGORY_NAMES


class PyTorchObjectDetectionOperator(Operator):
    """
    PyTorch object detection model operator, detect the region of interests in one image and predict the object classes
    in these regions.
    """

    def __init__(self, model_name) -> None:
        super().__init__()
        model_func = getattr(torchvision.models.detection, model_name)
        self._model = model_func(pretrained=True)
        self._model.eval()

    def __call__(self, images: List[torch.Tensor], mask_score=0.5) -> NamedTuple('Outputs', [('boxes', numpy.ndarray),
                                                                           ('breed', str)]):
        Outputs = NamedTuple('Outputs', [('boxes', numpy.ndarray), ('breed', str)])
        outputs = self._model(images)
        pred = outputs[0]
        scores = pred['scores']
        mask = scores >= mask_score
        boxes = pred['boxes'][mask]
        labels = pred['labels'][mask]
        if labels is not None:
            breeds = [lb_names[i] for i in labels] if lb_names is not None else None

        return Outputs(boxes.detach().numpy(), breeds)

    def train(self):
        """
        For training model
        """
        pass



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from typing import Dict
import threading
from queue import Queue

from towhee.dataframe import DataFrame
from towhee.engine.operator_io.reader import BatchFrameReader
from towhee.engine.operator_runner.runner_base import RunnerStatus
from towhee.engine.operator_runner.window_runner import WindowRunner
from towhee.tests.mock_operators.sum_operator.sum_operator import SumOperator


DATA_QUEUE = Queue()


class MockWriter:
    def __init__(self):
        self.res = []

    def write(self, data: Dict):
        self.res.append(data)


def run(runner):
    runner.process()


class TestMapRunner(unittest.TestCase):
    """
    MapRunner test
    """

    def test_flatmap_runner(self):
        writer = MockWriter()

        df_in = DataFrame('op_test_in', {'num': {'type': 'int', 'index': 0}})
        test = {df_in.name: {'df': df_in, 'cols': [('num', 0)]}}
        input_order = ['num']


        # We
        runner = WindowRunner('window_test', 0, 'sum_operator',
                              'mock_operators', {},
                              BatchFrameReader(test, input_order, 5, 3),
                              writer)

        runner.set_op(SumOperator())
        t = threading.Thread(target=run, args=(runner, ))
        t.start()
        self.assertEqual(runner.status, RunnerStatus.RUNNING)
        for _ in range(100):
            df_in.put_dict({'num': 1})
        df_in.seal()
        runner.join()
        self.assertEqual(runner.status, RunnerStatus.FINISHED)
        self.assertEqual(len(writer.res), 68)
        count = 0
        for item in writer.res:
            if count < 64:
                self.assertEqual(item.sum, 5)
            elif count < 66:
                self.assertEqual(item.sum, 4)
            else:
                self.assertEqual(item.sum, 1)
            count += 1

    def test_window_runner_with_error(self):
        writer = MockWriter()

        df_in = DataFrame('op_test_in', {'num': {'type': 'int', 'index': 0}})
        test = {df_in.name: {'df': df_in, 'cols': [('num', 0)]}}
        input_order = ['num']

        # We
        runner = WindowRunner('window_test', 0, 'sum_operator',
                              'mock_operators', {},
                              BatchFrameReader(test, input_order, 5, 3),
                              writer)

        runner.set_op(SumOperator())
        t = threading.Thread(target=run, args=(runner, ))
        t.start()
        self.assertEqual(runner.status, RunnerStatus.RUNNING)
        df_in.put(('error_data',))
        df_in.seal()
        runner.join()
        self.assertEqual(runner.status, RunnerStatus.FAILED)

if __name__ == '__main__':
    unittest.main()




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Iterable, List, Tuple, Any

from towhee.array import Array


class DataFrame:
    """
    A `DataFrame` is a collection of immutable, potentially heterogeneous blogs of data.

    Args:
        name (`str`):
            Name of the dataframe; `DataFrame` names should be the same as its
            representation.
        data (`list[towhee.Array]` or `list[Tuple]` or `dict[str, towhee.Array]`):
            The data of the `DataFrame`. Internally, the data will be organized
            in a column-based manner.
    """

    def __init__(
        self,
        name: str = None,
        data=None,
        columns=None,
    ):
        self._name = name
        self._len = 0
        self._sealed = False
        self._lock = threading.Lock()

        # For `data` is empty
        if not data:
            pass

        # For `data` is `list`
        elif isinstance(data, list):
            container_types = set(type(i) for i in data)
            if len(container_types) != 1:
                raise ValueError(
                    'can not construct Dataframe from a list of hybrid data containers. Try list[Tuple] or list[Array].')
            container_type = container_types.pop()

            # For `data` is `list[tuple]`
            if container_type is tuple:
                self._from_tuples(data, columns)
            # For `data` is `list[towhee.Array]`
            elif container_type is Array:
                self._from_arrays(data, columns)
            else:
                raise ValueError('can not construct DataFrame from list[%s]' % (container_type))

        # For `data` is `dict`
        elif isinstance(data, dict):
            self._from_dict(data)

        # Unrecognized data types
        else:
            raise ValueError('can not construct DataFrame from data type %s' % (type(data)))

    def __getitem__(self, key):
        # access a row
        if isinstance(key, int):
            return tuple(self._data_as_list[i][key] for i in range(len(self._data_as_list)))
        # access a column
        elif isinstance(key, str):
            return self._data_as_dict[key]

    def __len__(self):
        return self._len

    @property
    def name(self) -> str:
        return self._name

    @property
    def data(self) -> List[Array]:
        return self._data_as_list

    def iter(self) -> Iterable[Tuple[Any, ...]]:
        """
        Iterate over DataFrame rows as tuples.
        """
        return DFIterator(self)

    def seal(self):
        with self._lock:
            self._sealed = True

    def is_sealed(self) -> bool:
        with self._lock:
            return self._sealed

    def _from_tuples(self, data, columns):

        # check tuple length
        tuple_lengths = set(len(i) for i in data)
        if len(tuple_lengths) == 1:
            tuple_length = tuple_lengths.pop()
        else:
            raise ValueError('can not construct DataFrame from unequal-length tuples')

        # check columns length
        if columns and len(columns) != tuple_length:
            raise ValueError('length of columns is not equal to the length of tuple')

        # create arrays
        if columns:
            self._data_as_list = [Array(name=columns[i]) for i in range(tuple_length)]
            self._data_as_dict = {columns[i]: self._data_as_list[i] for i in range(tuple_length)}
        else:
            self._data_as_list = [Array()] * tuple_length
            self._data_as_dict = None

        # tuples to arrays
        for row in data:
            for i, element in enumerate(row):
                self._data_as_list[i].put(element)

        self._len = len(data)

    def _from_arrays(self, data, columns):

        # check array length
        array_lengths = set(len(array) for array in data)
        if len(array_lengths) != 1:
            raise ValueError('arrays in data should have equal length')

        self._len = array_lengths.pop()

        # check columns length
        if columns and len(columns) != len(data):
            raise ValueError('length of columns is not equal to the number of arrays')

        self._data_as_list = data

        if columns:
            self._data_as_dict = {columns[i]: self._data_as_list[i] for i in range(len(data))}
        else:
            self._data_as_dict = None

    def _from_dict(self, data):

        # check dict values
        for value in data.values():
            if not isinstance(value, Array):
                raise ValueError('value type in data should be towhee.Array')

        # check arrays length
        array_lengths = set(len(array) for array in data.values())
        if len(array_lengths) != 1:
            raise ValueError('arrays in data should have equal length')

        self._len = array_lengths.pop()

        self._data_as_list = list(data.values())
        self._data_as_dict = data


class DFIterator:
    """
    A row-based `DataFrame` iterator.
    """

    def __init__(self, df: DataFrame):
        self._df = df
        self._offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        """
        Returns:
            (`Tuple[Any, ...]`)
                In the normal case, the iterator will return a `Tuple` at each call.
            (`None`)
                In the case that the `DataFrame` is not sealed and the new rows are
                not ready yet, the iterator will return `None`. The caller should
                determine whether to block the iteration or exit the loop.
        Raises:
            (`StopIteration`)
                The iteration end iff the `DataFrame` is sealed and the last row is
                reached.
        """

        if len(self._df) == self._offset:
            if self._df.is_sealed():
                # Already reach the last row
                raise StopIteration
            else:
                # No more ready rows
                return None
        else:
            row = self._df[self._offset]
            self._offset += 1
            return row

    def ack(self):
        """
        To notice the DataFrame that the iterated rows has been successfully processed.

        An acknowledgement (ack) will notice the `DataFrame`s that the rows already
        iterated over are no longer used, and can be deleted from the system.
        """
        pass



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Tuple
from towhee.engine.status import Status
from abc import ABC
from enum import Enum, auto
from threading import Event
import traceback

from towhee.utils.log import engine_log


class RunnerStatus(Enum):
    IDLE = auto()
    RUNNING = auto()
    FINISHED = auto()
    FAILED = auto()

    @staticmethod
    def is_end(status: 'RunnerStatus') -> bool:
        return status in [RunnerStatus.FINISHED, RunnerStatus.FAILED]


class _OpInfo:
    """
    Operator hub info.
    """

    def __init__(self, op_name: str, hub_op_id: str,
                 op_args: Dict[str, any]) -> None:
        self._op_name = op_name
        self._hub_op_id = hub_op_id
        self._msg = None
        self._op_args = op_args

    @property
    def op_name(self):
        return self._op_name

    @property
    def hub_op_id(self):
        return self._hub_op_id

    @property
    def op_args(self):
        return self._op_args


class RunnerBase(ABC):
    """
    Running instance of op.

    The base class provides some function to control status.
    """

    def __init__(self, name: str, index: int,
                 op_name: str, hub_op_id: str,
                 op_args: Dict[str, any],
                 reader=None, writer=None) -> None:
        self._name = name
        self._index = index
        self._status = RunnerStatus.IDLE
        self._op_info = _OpInfo(op_name, hub_op_id, op_args)
        self._reader = reader
        self._writer = writer
        self._need_stop = False
        self._end_event = Event()
        self._op = None

    def _set_end_status(self, status: RunnerStatus):
        self._set_status(status)
        self._end_event.set()
        engine_log.info('%s ends with status: %s', self._name, status)

    @property
    def op_name(self):
        return self._op_info.op_name

    @property
    def hub_op_id(self):
        return self._op_info.hub_op_id

    @property
    def op_args(self):
        return self._op_info.op_args

    @property
    def status(self) -> RunnerStatus:
        return self._status

    @property
    def msg(self):
        return self._msg

    @property
    def op(self):
        return self._op

    def is_end(self) -> bool:
        return RunnerStatus.is_end(self.status)

    def set_op(self, op):
        self._op = op

    def unset_op(self):
        self._op = None

    def _set_finished(self) -> None:
        self._set_end_status(RunnerStatus.FINISHED)

    def _set_idle(self) -> None:
        self._set_status(RunnerStatus.IDLE)

    def _set_failed(self, msg: str) -> None:
        error_info = '{} runs failed, error msg: {}'.format(str(self), msg)
        engine_log.error(error_info)
        self._msg = error_info
        self._set_end_status(RunnerStatus.FAILED)

    def _set_running(self) -> None:
        self._set_status(RunnerStatus.RUNNING)

    def _set_status(self, status: RunnerStatus) -> None:
        self._status = status

    def is_idle(self) -> bool:
        return self.status == RunnerStatus.IDLE

    def is_finished(self) -> bool:
        return self.status == RunnerStatus.FINISHED

    def set_stop(self) -> None:
        engine_log.info('Begin to stop %s', str(self))
        self._need_stop = True
        self._reader.close()

    def __str__(self) -> str:
        return '{}:{}'.format(self._name, self._index)

    def _call_op(self, inputs) -> Status:
        try:
            outputs = self._op(**inputs)
            return Status.ok_status(outputs)
        except Exception as e:  # pylint: disable=broad-except
            err = '{}, {}'.format(str(e), traceback.format_exc())
            return Status.err_status(err)

    def _get_inputs(self) -> Tuple[bool, Dict[str, any]]:
        try:
            data = self._reader.read()
            return False, data
        except StopIteration:
            return True, None

    def _set_outputs(self, output: any):
        self._writer.write(output)

    def process_step(self) -> bool:
        is_end, op_input_params = self._get_inputs()
        if is_end:
            self._set_finished()
            return True

        if op_input_params is None:
            # No data in dataframe, but dataframe is not sealed
            self._set_idle()
            return True

        st = self._call_op(op_input_params)
        if st.is_ok():
            self._set_outputs(st.data)
            return False
        else:
            self._set_failed(st.msg)
            return True

    def join(self):
        self._end_event.wait()

    def process(self):
        engine_log.info('Begin to run %s', str(self))
        self._set_running()
        while True:
            if not self._need_stop:
                try:
                    if self.process_step():
                        break
                except Exception as e:  # pylint: disable=broad-except
                    err = '{}, {}'.format(e, traceback.format_exc())
                    self._set_failed(err)
                    break
            else:
                self._set_finished()
                break



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
from pathlib import Path
from typing import Any, Dict

from towhee.operator import Operator
from towhee.operator.nop import NOPOperator
from towhee.engine import LOCAL_OPERATOR_CACHE
from towhee.hub.file_manager import FileManager


class OperatorLoader:
    """Wrapper class used to load operators from either local cache or a remote
    location.

    Args:
        cache_path: (`str`)
            Local cache path to use. If not specified, it will default to
            `$HOME/.towhee/operators`.
    """
    def __init__(self, cache_path: str = None):
        if cache_path is None:
            self._cache_path = LOCAL_OPERATOR_CACHE
        else:
            self._cache_path = Path(cache_path)

    def load_operator(self, function: str, args: Dict[str, Any]) -> Operator:
        """Attempts to load an operator from cache. If it does not exist, looks up the
        operator in a remote location and downloads it to cache instead. By standard
        convention, the operator must be called `Operator` and all associated data must
        be contained within a single directory.

        Args:
            function: (`str`)
                Origin and method/class name of the operator. Used to look up the proper
                operator in cache.
        Raises:
            FileExistsError
                Cannot find operator.
        """

        if function in ['_start_op', '_end_op']:
            return NOPOperator()

        fm = FileManager()
        path = fm.get_operator(function)

        if path is None:
            raise FileExistsError('Cannot find operator.')

        fname = str(path).rsplit('/', maxsplit=1)[-1][:-3]
        modname = 'towhee.operator.' + fname

        spec = importlib.util.spec_from_file_location(modname, path.resolve())

        # Create the module and then execute the module in its own namespace.
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        # Instantiate the operator object and return it to the caller for
        # `load_operator`. By convention, the operator class is simply the CamelCase
        # version of the snake_case operator.
        op_cls = ''.join(x.capitalize() or '_' for x in fname.split('_'))
        if args is not None:
            return getattr(module, op_cls)(**args)
        else:
            return getattr(module, op_cls)()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import time
import queue

from towhee.dataframe import DataFrame, DataFrameIterator

from towhee.tests.test_util.dataframe_test_util import DfWriter, MultiThreadRunner


class TestDataframe(unittest.TestCase):
    """
    Test dataframe basic function
    """

    def test_put(self):
        df = DataFrame('test')
        t = DfWriter(df, 10)
        t.start()
        t.join()
        self.assertEqual(df.name, 'test')
        self.assertEqual(df.size, 10)
        datas = df.get(0, 4)
        self.assertEqual(len(datas), 4)

        datas = df.get(8, 4)
        self.assertEqual(datas, None)

        self.assertFalse(df.sealed)
        df.seal()
        self.assertTrue(df.sealed)

        datas = df.get(8, 4)
        self.assertEqual(len(datas), 2)

    def test_multithread(self):
        df = DataFrame('test')
        data_size = 10
        t = DfWriter(df, data_size)
        t.set_sealed_when_stop()
        t.start()
        q = queue.Queue()

        def read(df: DataFrame, q: queue.Queue):
            index = 0
            while True:
                items = df.get(index, 2)
                if items:
                    for item in items:
                        q.put(item)
                        index += 1
                if df.sealed:
                    items = df.get(index, 100)
                    if items:
                        for item in items:
                            q.put(item)
                    break

        runner = MultiThreadRunner(target=read, args=(df, q), thread_num=10)

        runner.start()
        runner.join()

        self.assertEqual(q.qsize(), data_size * 10)

    def test_multithread_block_get(self):
        df = DataFrame('test')
        data_size = 10
        t = DfWriter(df, data_size)
        t.set_sealed_when_stop()
        t.start()
        q = queue.Queue()

        def read(df: DataFrame, q: queue.Queue):
            index = 0
            while True:
                items = df.get(index, 2, True)
                if items:
                    for item in items:
                        q.put(item)
                        index += 1
                    continue
                if df.sealed:
                    items = df.get(index, 100)
                    if items:
                        for item in items:
                            q.put(item)
                    break

        runner = MultiThreadRunner(target=read, args=(df, q), thread_num=10)

        runner.start()
        runner.join()

        self.assertEqual(q.qsize(), data_size * 10)

    def test_notify(self):
        df = DataFrame('test')

        def read(df: DataFrame):
            index = 0
            while True:
                item = df.get(index, 2, True)
                assert item is None
                break

        runner = MultiThreadRunner(target=read, args=(df, ), thread_num=10)
        runner.start()
        time.sleep(0.02)

        # if not call notify_block_readers, reader thread will never return
        df.notify_block_readers()
        runner.join()


class TestMapIterator(unittest.TestCase):
    """
    map iterator basic test
    """

    def test_map_iterator(self):
        df = DataFrame('test')
        it = df.map_iter()
        data_size = 100
        t = DfWriter(df, data_size)
        t.start()
        t.set_sealed_when_stop()
        q = queue.Queue()

        def read(it: DataFrameIterator, q: queue.Queue):
            for item in it:
                if item is not None:
                    q.put(item)
                time.sleep(0.01)

        runner = MultiThreadRunner(target=read, args=(it, q), thread_num=1)
        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 100)

    def test_block_map_iterator(self):
        df = DataFrame('test')
        it = df.map_iter(True)
        data_size = 100
        t = DfWriter(df, data_size)
        t.start()
        t.set_sealed_when_stop()
        q = queue.Queue()

        def read(it: DataFrameIterator, q: queue.Queue):
            for item in it:
                if item is not None:
                    q.put(item)
                time.sleep(0.01)

        runner = MultiThreadRunner(target=read, args=(it, q), thread_num=1)
        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 100)

    def test_notify(self):
        df = DataFrame('test')
        it = df.map_iter()

        def read(it: DataFrameIterator):
            for item in it:
                assert item is None
                break

        runner = MultiThreadRunner(target=read, args=(it, ), thread_num=5)
        runner.start()
        time.sleep(0.5)
        it.notify()
        runner.join()


class TestBatchIter(unittest.TestCase):
    """
    Batch iter test.
    """

    def test_batch_iterator(self):
        df = DataFrame('test')
        it = df.batch_iter(3, 4)
        data_size = 100
        t = DfWriter(df, data_size)
        t.start()
        t.set_sealed_when_stop()
        q = queue.Queue()

        def read(it: DataFrameIterator, q: queue.Queue):
            for item in it:
                if item is not None:
                    q.put(item)
                time.sleep(0.01)

        runner = MultiThreadRunner(target=read, args=(it, q), thread_num=1)
        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 25)
        while not q.empty():
            item = q.get()
            self.assertEqual(item, [(), (), ()])

    def test_block_batch_iterator(self):
        df = DataFrame('test')
        it = df.batch_iter(4, 2, True)
        data_size = 100
        t = DfWriter(df, data_size)
        t.start()
        t.set_sealed_when_stop()
        q = queue.Queue()

        def read(it: DataFrameIterator, q: queue.Queue):
            for item in it:
                if item is not None:
                    q.put(item)
                time.sleep(0.01)

        runner = MultiThreadRunner(target=read, args=(it, q), thread_num=1)
        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 50)
        num = 0
        while not q.empty():
            item = q.get()
            if num < 49:
                self.assertEqual(item, [(), (), (), ()])
            else:
                self.assertEqual(item, [(), ()])
            num += 1


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import cv2
from typing import Dict


def get_video_info(filename: str) -> Dict:
    """
    get properties of video file.

    Args:
        filename(str):
            A video file.
    Returns:
        (Dict):
            A dict with all the video properties.
    """
    cap = cv2.VideoCapture(filename)
    if cap.isOpened():
        fps = int(round(cap.get(cv2.CAP_PROP_FPS)))
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        duration = frame_count / fps
        cap.release()
        video_info = {
            'fps': fps,
            'width': width,
            'height': height,
            'frame_count': frame_count,
            'duration': duration
        }
        return video_info
    else:
        raise ValueError(f'OpenCV failed to open {filename}.')



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import sys
import os

from torchvision import transforms
from towhee.data.dataset.image_datasets import PyTorchImageDataset

cache_path = os.path.abspath(sys.path[0])
image_path = os.path.join(cache_path, 'kaggle_dataset_small/train')
label_file = os.path.join(cache_path, 'kaggle_dataset_small/train/train_labels.csv')


class DatasetTest(unittest.TestCase):
    def test_length(self):
        data_transforms = {
            'train': transforms.Compose([
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            ])
        }
        dataset = PyTorchImageDataset(image_path, label_file, data_transforms['train'])
        images, labels = dataset.images, dataset.labels
        self.assertEqual(len(images), len(labels))


if __name__ == '__main__':
    unittest.main(verbosity=1)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Any, Callable, List, Tuple, Dict
import weakref

from towhee.dataframe.variable import Variable


class DataFrame:
    """A dataframe is a collection of immutable, potentially heterogeneous blogs of
    data.
    """

    def __init__(self, name: str = None, cols=None, data: List[Tuple[Variable]] = None):
        """DataFrame constructor.

        Args:
            name:
                Name of the dataframe; `DataFrame` names should be the same as its
                representation.
            data:
                A list of data tuples - in all instances, the number of elements per
                tuple should be identical throughout the entire lifetime of the
                `Dataframe`. These tuples can be interpreted as being direct outputs
                into downstream operators.
        """
        self._name = name
        # TODO (junjie.jiangjjj) define col struct
        self._cols = cols
        self._data = data if data else []

        # `_start_idx` corresponds to the actual index for the current 0th element in
        # this `DataFrame`.
        self._start_idx = 0
        self._total = len(self._data)

        # A list of `DataFrameIterator` instances registered to this `DataFrame`.
        self._iters = []

        # Sealed `DataFrames` are disabled, while changes made to `DataFrame` state must
        # first acquire this lock. This flag and condition variable track whether or not
        # this `DataFrame` can still be used.
        self._sealed = False
        self._lock = threading.Lock()
        self._seal_cv = threading.Condition(self._lock)
        self._accessible_cv = threading.Condition(self._lock)

    @property
    def name(self) -> str:
        return self._name

    @property
    def data(self) -> List[Tuple[Variable]]:
        return self._data

    @property
    def size(self) -> int:
        return self._total

    @property
    def sealed(self) -> bool:
        """
        Sealed `DataFrame`s are disabled.
        """
        return self._sealed

    def put(self, item: Tuple[Variable]) -> None:
        assert not self._sealed, f'DataFrame {self._name} is already sealed, can not put data'
        assert isinstance(item, tuple), 'Dataframe needs tuple, not %s' % (type(item))
        with self._lock:
            self._data.append(item)
            self._total += 1
            self._accessible_cv.notify()

    def put_dict(self, data: Dict[str, Any]):
        datalist = [None] * len(self._cols)
        for k, v in data.items():
            datalist[self._cols[k]['index']] = Variable(
                self._cols[k]['type'], v)
        self.put(tuple(datalist))

    def merge(self, df):
        assert not self._sealed, f'DataFrame {self._name} is already sealed, can not put data'
        # TODO: Check `df` compatibility with `self`.
        with self._lock:
            self._data += df.data
            self._total += len(df.data)
            self._accessible_cv.notify_all()

    def get(self, start: int, count: int, block: bool = False) -> List[Tuple[Variable]]:
        """
        Get [start: start + count) elements within the `DataFrame`.

        If there are enough elements within this `DataFrame`, this function will gather
        the elements and return them. If this `DataFrame` is already sealed and fewer
        than `count` elements are available, this function will return all the remaining
        elements. In all other cases, this function will return `None`.

        Args:
            start: (`int`)
                The start index within the `DataFrame`.
            count: (`int`)
                Number of elements to acquire.

        Returns:
            data: (`List[Tuple[Variable]]`)
                At most `count` elements acquired from the `DataFrame`, or `None` if
                not enough elements are available.
        """

        # TODO (junjie.jiang)
        # 1. call gc function
        # 2. put and seal will be called by the same operator
        # 3. seal will always be called after put
        # 4. iterators never ends until they meet df._sealed == True

        with self._accessible_cv:
            if start < self._start_idx:
                raise IndexError(
                    'Can not read from {start}, dataframe {name} start index is {cur_start}',
                    start=start,
                    name=self._name,
                    cur_start=self._start_idx
                )

            if block and not self._accessible(start, count):
                self._accessible_cv.wait()

            if self._total - start >= count:
                idx0 = start - self._start_idx
                return self._data[idx0:idx0+count]

            # If the `DataFrame` is already sealed, return only the remaining data.
            if self._sealed:
                if self._total <= start:
                    return None
                return self._data[start:]
            return None

    def notify_block_readers(self):
        with self._accessible_cv:
            self._accessible_cv.notify_all()

    def _accessible(self, start: int, count: int) -> bool:
        if self._sealed:
            return True
        if self._total - start >= count:
            return True
        return False

    def seal(self):
        with self._seal_cv:
            self._sealed = True
            self._seal_cv.notify_all()
            self._accessible_cv.notify_all()

    def wait_sealed(self):
        with self._seal_cv:
            while not self._sealed:
                self._seal_cv.wait()

    def clear(self):
        with self._lock:
            self._data = []
            self._start_idx = 0
            self._total = 0
            self._sealed = False

    def _gc(self) -> None:
        """
        Delete the data which all registered iter has read
        """
        # TODO (junjie.jiangjjj)
        raise NotImplementedError

    def map_iter(self, block: bool = False):
        it = MapIterator(self, block)
        with self._lock:
            self._iters.append(it)
        return it

    def batch_iter(self, size: int, step: int, block: bool = False):
        it = BatchIterator(self, size, step, block)
        with self._lock:
            self._iters.append(it)
        return it

    def __str__(self) -> str:
        return 'DataFrame [%s] with [%s] datas, seal: [%s]' % (self._name, self.size, self._sealed)


class DataFrameIterator:
    """Base iterator implementation. All iterators should be subclasses of
    `DataFrameIterator`.

    Args:
        df: The dataframe to iterate over.
    """

    def __init__(self, df: DataFrame):
        self._df_ref = weakref.ref(df)
        self._cur_idx = 0
        self._df_name = df.name

    def __iter__(self):
        return self

    def __next__(self) -> List[Tuple]:
        # The base iterator is purposely defined to have exatly 0 elements.
        raise StopIteration

    @property
    def df_name(self) -> int:
        """Returns the df name.
        """
        return self._df_name

    @property
    def current_index(self) -> int:
        """Returns the current index.
        """
        return self._cur_idx

    @property
    def accessible_size(self) -> int:
        """Returns current accessible data size.
        """
        return self._df_ref().size - self._cur_idx

    def notify(self):
        df = self._df_ref()
        if df is not None:
            self._df_ref().notify_block_readers()


class MapIterator(DataFrameIterator):
    """Iterator implementation that traverses the dataframe line-by-line.

    Args:
        df: The dataframe to iterate over.
    """

    def __init__(self, df: DataFrame, block: bool = False):
        super().__init__(df)
        self._block = block

    def __next__(self) -> List[Tuple]:
        data = self._df_ref().get(self._cur_idx, 1, self._block)
        if self._df_ref().sealed and not data:
            raise StopIteration
        if data is not None:
            self._cur_idx += 1
            return data[0]
        return None


class BatchIterator(DataFrameIterator):
    """Iterator implementation that traverses the dataframe multiple rows at a time.

    Args:
        df:
            The dataframe to iterate over.
        size:
            batch size. If size is None, then the whole variable set will be batched
            together.
    """

    def __init__(self, df: DataFrame, size: int, step: int, block=False):
        super().__init__(df)
        self._size = size
        self._step = step
        self._block = block

    def __next__(self):
        data = None
        while data is None:
            data = self._df_ref().get(self._cur_idx, self._size, self._block)
            if self._df_ref().sealed and not data:
                raise StopIteration

            if data is not None:
                self._cur_idx += self._step
                return data

            elif not self._block:
                return None


class GroupIterator:
    """Iterator implementation that traverses the dataframe based on a custom group-by
    function.

    Args:
        df:
            The dataframe to iterate over.
        func:
            The function used to return grouped data within the dataframe.
    """

    def __init__(self, df: DataFrame, func: Callable[[], None]):
        super().__init__(df)
        self._func = func

    def __next__(self):
        raise NotImplementedError


class RepeatIterator:
    """Iterator implementation that repeatedly accesses the first line of the dataframe.

        Args:
            df:
                The dataframe to iterate over.
            n:
                the repeat times. If `None`, the iterator will loop continuously.
    """

    def __init__(self, df: DataFrame, n: int = None):
        # self._n = n
        # self._i = 0
        raise NotImplementedError

    def __next__(self):
        if self._i >= self._n:
            raise StopIteration
        return self._data[0]



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
import torch
from towhee.models.layers.padding_functions import get_padding, get_same_padding, is_static_pad, pad_same, get_padding_value

class TestPaddingFunctions(unittest.TestCase):
    """
    Test Padding utility functions
    """
    def test_get_padding(self):
        """
        Test get_padding_function.
        """
        kernel_size = 3
        stride = 1
        dilation = 1
        self.assertEqual(get_padding(kernel_size, stride, dilation), 1)

        kernel_size = 3
        stride = 2
        dilation = 2
        self.assertEqual(get_padding(kernel_size, stride, dilation), 2)

    def test_get_same_padding(self):
        """
        Test get_same_padding function.
        """
        x = 224
        k = 3
        s = 1
        d = 1
        self.assertEqual(get_same_padding(x,k,s,d), 2)

        x = 224
        k = 5
        s = 1
        d = 1
        self.assertEqual(get_same_padding(x,k,s,d), 4)

    def test_is_static_pad(self):
        """
        Test get_same_padding function.
        """
        kernel = 3
        stride = 1
        dilation = 1
        self.assertEqual(is_static_pad(kernel, stride, dilation), True)

        kernel = 3
        stride = 1
        dilation = 2
        self.assertEqual(is_static_pad(kernel, stride, dilation), True)

    def test_get_padding_value(self):
        """
        Test get_padding_value function.
        """
        kernel = 3
        stride = 1
        dilation = 1

        padding, dynamic = get_padding_value('same', kernel, stride = stride, dilation = dilation)
        self.assertEqual(padding, 1)
        self.assertEqual(dynamic, False)

    def test_pad_same(self):
        """
        Test pad same function.
        """
        x = torch.Tensor(1,3,224,224)
        k = [3,3]
        s = [1,1]
        d = [1,1]
        value = 128.0
        out = pad_same(x, k, s, d, value)
        self.assertEqual(out.shape, torch.Size((1,3,226,226)))
        self.assertEqual(out[0,0,0,0], 128.0)




from .weight_init import trunc_normal_, variance_scaling_, lecun_normal_
from .mlp import Mlp
from .convmlp import ConvMlp
from .gatedmlp import GatedMlp
from .dropblock2d import DropBlock2d, drop_block_2d
from .droppath import DropPath





# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from enum import Enum, auto

from towhee.dag.operator_repr import OperatorRepr
from towhee.engine.operator_runner.runner_base import RunnerStatus
from towhee.dataframe import DataFrame
from towhee.engine.operator_io import create_reader, create_writer
from towhee.engine.operator_runner import create_runner
from towhee.engine.thread_pool_task_executor import ThreadPoolTaskExecutor


class OpStatus(Enum):
    NOT_RUNNING = auto()
    RUNNING = auto()
    FINISHED = auto()
    FAILED = auto()
    STOPPED = auto()


class OperatorContext:
    """
    The OperatorContext manages an operator's input data and output data at runtime,
    as well as the operators' dependency within a GraphContext.
    The abstraction of OperatorContext hides the complexity of Dataframe management,
    input iteration, and data dependency between Operators. It offers a Task-based
    scheduling context.

    Args:
        op_repr: (OperatorRepr)
            The operator representation
        dataframes: (`dict` of `DataFrame`)
            All the `DataFrames` in `GraphContext`
    """
    def __init__(self, op_repr: OperatorRepr, dataframes: Dict[str, DataFrame]):
        self._repr = op_repr
        iter_type = op_repr.iter_info['type']
        reader_inputs = {}
        input_order = []
        for x in op_repr.inputs:
            # If a dataframe is already added, append the new columns.
            if x['df'] in reader_inputs.keys():
                reader_inputs[x['df']]['cols'].append((x['name'], x['col']))
                input_order.append(x['name'])
            else:
                reader_inputs[x['df']] = {}
                reader_inputs[x['df']]['df'] = dataframes[x['df']]
                reader_inputs[x['df']]['cols'] = [(x['name'], x['col'])]
                input_order.append(x['name'])

        self._reader = create_reader(reader_inputs, input_order, iter_type)

        outputs = list({dataframes[output['df']] for output in op_repr.outputs})

        self._writer = create_writer(iter_type, outputs)
        self._op_runners = []

        self._op_status = OpStatus.NOT_RUNNING
        self._err_msg = None

    @property
    def name(self):
        return self._repr.name

    @property
    def err_msg(self):
        return self._err_msg

    @property
    def status(self):
        """
        Calc op-ctx status by checking all runners of this op-ctx
        """
        if self._op_status in [OpStatus.FINISHED, OpStatus.FAILED]:
            return self._op_status

        if len(self._op_runners) == 0:
            return self._op_status

        finished_count = 0
        for runner in self._op_runners:
            if runner.status == RunnerStatus.FAILED:
                self._op_status = OpStatus.FAILED
                self._err_msg = runner.msg
            else:
                if runner.status == RunnerStatus.FINISHED:
                    finished_count += 1
        if finished_count == len(self._op_runners):
            self._op_status = OpStatus.FINISHED
        return self._op_status

    def start(self, executor: ThreadPoolTaskExecutor, count: int = 1) -> None:
        if self._op_status != OpStatus.NOT_RUNNING:
            raise RuntimeError('OperatorContext can only be started once')

        self._op_status = OpStatus.RUNNING

        for i in range(count):
            self._op_runners.append(
                create_runner(
                    self._repr.iter_info['type'],
                    self._repr.name,
                    i,
                    self._repr.name,
                    self._repr.function,
                    self._repr.init_args,
                    self._reader,
                    self._writer
                )
            )
        for runner in self._op_runners:
            executor.push_task(runner)

    def stop(self):
        if self.status != OpStatus.RUNNING:
            raise RuntimeError('Op ctx is already stopped.')

        for runner in self._op_runners:
            runner.set_stop()

    def join(self):
        # Waits all runner finished.
        for runner in self._op_runners:
            runner.join()
        self._writer.close()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import NamedTuple, List

import cv2
from PIL import Image

from towhee.operator import Operator, SharedType


class Cv2Decoder(Operator):
    """
    Cv2 decoder
    """

    def __init__(self):
        super().__init__()
        self.key = ()

    def __call__(self, video_path: str) -> NamedTuple('Outputs', [('imgs', List['Image'])]):
        Outputs = NamedTuple('Outputs', [('imgs', List['Image'])])
        imgs = []
        cap = cv2.VideoCapture(video_path)
        while True:
            _, frame = cap.read()
            if frame is not None:
                imgs.append(Image.fromarray(frame))
            else:
                cap.release()
                return Outputs(imgs)

    @property
    def shared_type(self):
        return SharedType.Shareable



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from pathlib import Path
# from shutil import rmtree

from towhee.operator import Operator
from towhee.engine.operator_pool import OperatorPool
from towhee.engine.operator_runner.runner_base import _OpInfo
from towhee.tests import CACHE_PATH
from towhee.hub.file_manager import FileManagerConfig, FileManager


class TestOperatorPool(unittest.TestCase):
    """Basic test case for `OperatorPool`.
    """
    @classmethod
    def setUpClass(cls):
        new_cache = (CACHE_PATH/'test_cache')
        pipeline_cache = (CACHE_PATH/'test_util')
        operator_cache = (CACHE_PATH/'mock_operators')
        fmc = FileManagerConfig()
        fmc.update_default_cache(new_cache)
        pipelines = list(pipeline_cache.rglob('*.yaml'))
        operators = [f for f in operator_cache.iterdir() if f.is_dir()]
        fmc.cache_local_pipeline(pipelines)
        fmc.cache_local_operator(operators)
        FileManager(fmc)

    # @classmethod
    # def tearDownClass(cls):
    #     new_cache = (CACHE_PATH/'test_cache')
    #     rmtree(str(new_cache))

    def setUp(self):
        cache_path = Path(__file__).parent.parent.resolve()
        self._op_pool = OperatorPool(cache_path=cache_path)

    def test_acquire_release(self):
        hub_op_id = 'local/add_operator'
        op_info = _OpInfo('add_operator', hub_op_id, {'factor': 0})

        op = self._op_pool.acquire_op(op_info.hub_op_id,
                                      op_info.op_args)

        # Perform some simple operations.
        self.assertTrue(isinstance(op, Operator))
        self.assertEqual(op(1).sum, 1)

        # Release and re-acquire the operator.
        self._op_pool.release_op(op)
        op = self._op_pool.acquire_op(op_info.hub_op_id,
                                      op_info.op_args)
        # Perform more operations.
        self.assertEqual(op(-1).sum, -1)
        self.assertEqual(op(100).sum, 100)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT_ WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from towhee.engine.operator_context import OperatorContext, OpStatus
from towhee.dataframe import DataFrame, Variable
from towhee.dag import OperatorRepr
from towhee.hub.file_manager import FileManagerConfig, FileManager
from towhee.engine.thread_pool_task_executor import ThreadPoolTaskExecutor
from towhee.tests.test_util.dataframe_test_util import DfWriter
from towhee.tests import CACHE_PATH


class TestOperatorContext(unittest.TestCase):
    """
    Op ctx test
    """

    @classmethod
    def setUpClass(cls):
        new_cache = (CACHE_PATH/'test_cache')
        pipeline_cache = (CACHE_PATH/'test_util')
        operator_cache = (CACHE_PATH/'mock_operators')
        fmc = FileManagerConfig()
        fmc.update_default_cache(new_cache)
        pipelines = list(pipeline_cache.rglob('*.yaml'))
        operators = [f for f in operator_cache.iterdir() if f.is_dir()]
        fmc.cache_local_pipeline(pipelines)
        fmc.cache_local_operator(operators)
        FileManager(fmc)

    def setUp(self):
        self._task_exec = ThreadPoolTaskExecutor('op_ctx_test_executor',
                                                 (CACHE_PATH/'mock_operators'))
        self._task_exec.start()

    def tearDown(self):
        if self._task_exec.is_alive():
            self._task_exec.stop()
            self._task_exec.join()

    def _create_op_ctx(self):
        df_in = DataFrame('op_test_in', {'num': {'type': 'int', 'index': 0}})
        df_out = DataFrame('op_test_out', {'sum': {'type': 'int', 'index': 0}})
        dfs = {'op_test_in': df_in, 'op_test_out': df_out}

        op_repr = OperatorRepr(
            name='add_operator',
            function='local/add_operator',
            init_args={'factor': 5},
            inputs=[
                {'name': 'num', 'df': 'op_test_in', 'col': 0},
            ],
            outputs=[{'df': 'op_test_out'}],
            iter_info={'type': 'map'}
        )

        op_ctx = OperatorContext(op_repr, dfs)
        self.assertEqual(op_ctx.name, 'add_operator')
        self.assertEqual(op_ctx.status, OpStatus.NOT_RUNNING)
        op_ctx.start(self._task_exec)
        self.assertEqual(op_ctx.status, OpStatus.RUNNING)
        return df_in, df_out, op_ctx

    def test_op_ctx(self):
        df_in, df_out, op_ctx = self._create_op_ctx()

        data = (Variable('int', 1), Variable(
            'str', 'test'), Variable('float', 0.1))
        data_size = 20
        t = DfWriter(df_in, data_size, data=data)
        t.set_sealed_when_stop()
        t.start()
        t.join()
        op_ctx.join()
        self.assertEqual(op_ctx.status, OpStatus.FINISHED)
        df_out.seal()
        map_iter = df_out.map_iter()
        for item in map_iter:
            self.assertEqual(item[0].value, 6)

    def test_op_ctx_failed(self):
        df_in, df_out, op_ctx = self._create_op_ctx()

        # Set errer data
        data = (Variable('str', 'test'), )
        data_size = 20
        t = DfWriter(df_in, data_size, data=data)
        t.set_sealed_when_stop()
        t.start()
        t.join()
        op_ctx.join()
        self.assertEqual(op_ctx.status, OpStatus.FAILED)
        df_out.seal()

        self.assertEqual(df_out.size, 0)

    def test_op_ctx_stop(self):
        df_in, _, op_ctx = self._create_op_ctx()

        data = (Variable('int', 1), Variable(
            'str', 'test'), Variable('float', 0.1))
        data_size = 20
        t = DfWriter(df_in, data_size, data=data)
        t.start()
        t.join()
        op_ctx.stop()
        op_ctx.join()
        self.assertEqual(op_ctx.status, OpStatus.FINISHED)

if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pathlib import Path

from .cv2_decoder import Cv2Decoder


class TestCv2Decoder(unittest.TestCase):
    """
    cv2 decoder test
    """

    def test_cv2_decoder(self):
        test_video = Path(__file__).parent / 'test_video.avi'
        decoder = Cv2Decoder()
        out = decoder(str(test_video))

        self.assertEqual(len(out.imgs), 85)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import unittest

from towhee.models.layers.vision_transformer_basic_head import VisionTransformerBasicHead


class VisionTransformerBasicHeadTest(unittest.TestCase):
    def test_vision_transformer_basic_head(self):
        batch_size = 8
        seq_len = 10
        input_dim = 10
        out_dim = 20
        head = VisionTransformerBasicHead(
            in_features=input_dim,
            out_features=out_dim,
        )
        fake_input = torch.rand(batch_size, seq_len, input_dim)
        output = head(fake_input)
        gt_shape = (batch_size, out_dim)
        self.assertEqual(tuple(output.shape), gt_shape)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
A collection of urls for open sourced models.
"""

from .vit_configs import (
    get_b16_config,
    get_b32_config,
    get_l16_config,
    get_l32_config,
    drop_head_variant
)


PRETRAINED_MODELS = {
    'B_16': {
      'config': get_b16_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_16.pth'
    },
    'B_32': {
      'config': get_b32_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_32.pth'
    },
    'L_16': {
      'config': get_l16_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': None
    },
    'L_32': {
      'config': get_l32_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/L_32.pth'
    },
    'B_16_imagenet1k': {
      'config': drop_head_variant(get_b16_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_16_imagenet1k.pth'
    },
    'B_32_imagenet1k': {
      'config': drop_head_variant(get_b32_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_32_imagenet1k.pth'
    },
    'L_16_imagenet1k': {
      'config': drop_head_variant(get_l16_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/L_16_imagenet1k.pth'
    },
    'L_32_imagenet1k': {
      'config': drop_head_variant(get_l32_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': 'https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/L_32_imagenet1k.pth'
    },
}



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import NamedTuple
from towhee.operator import Operator
from numpy import linalg as LA
import numpy


class NormDataOperator(Operator):
    """
    Stateful operator
    """

    # TODO: init without argument
    def __init__(self, normed_num) -> None:
        super().__init__()
        self.normed_num = normed_num

    def __call__(self, data: numpy.ndarray) -> NamedTuple("Outputs", [("norm_data", numpy.ndarray)]):
        norm_embedding = data[0] / LA.norm(data[0])
        Outputs = NamedTuple("Outputs", [("norm_data", numpy.ndarray)])
        return Outputs(norm_embedding)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from PIL import Image
from torchvision import transforms
import os

from towhee.trainer.models.vit.vit import ViT
from towhee.tests.mock_operators import PYTORCH_TRANSFORMER_VEC_OPERATOR_PATH, load_local_operator


class TransformerEmbeddingOperatorTest(unittest.TestCase):
    name = 'B_16_imagenet1k'
    pwd = os.getcwd()
    test_dir = pwd+'/towhee/tests/models/vit/'
    #weights_path = test_dir + 'B_16_imagenet1k.pth'
    weights_path = None
    model = ViT(name, weights_path=weights_path, pretrained=True, to_vec=True)
    print(f'model to_vec is {model.to_vec}')
    img = Image.open(test_dir + 'img.jpg')
    tfms = transforms.Compose([transforms.Resize(model.image_size), transforms.ToTensor(),
                               transforms.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5]), ])
    img = tfms(img).unsqueeze(0)

    def test_transformer_embedding_operator(self):
        trans = load_local_operator('pytorch_transformer_vec_operator', PYTORCH_TRANSFORMER_VEC_OPERATOR_PATH)
        op = trans.PytorchTransformerVecOperator(self.model)
        outputs = op(self.img)
        self.assertEqual(outputs.embedding.shape[0], 768)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz and Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import unittest
import numpy as np
from torch import nn
import itertools


from towhee.models.layers.resnet_basic_3d_module import ResNetBasic3DModule
from towhee.models.utils.create_resnet_basic_3d_module import create_resnet_basic_3d_module


class TestResNetBasic3DModule(unittest.TestCase):
    """
    Test ResNet basic 3d module.
    """
    def setUp(self):
        super().setUp()
        torch.set_rng_state(torch.manual_seed(42).get_state())

    @staticmethod
    def _get_inputs(input_dim: int = 3) -> torch.tensor:
        """
        Provide different tensors as test cases.
        Yield:
            (torch.tensor): tensor as test case input.
        """
        # Prepare random tensor as test cases.
        shapes = (
            # Forward succeeded.
            (1, input_dim, 3, 7, 7),
            (1, input_dim, 5, 7, 7),
            (1, input_dim, 7, 7, 7),
            (2, input_dim, 3, 7, 7),
            (4, input_dim, 3, 7, 7),
            (8, input_dim, 3, 7, 7),
            (2, input_dim, 3, 7, 14),
            (2, input_dim, 3, 14, 7),
            (2, input_dim, 3, 14, 14),
            # Forward failed.
            (8, input_dim * 2, 3, 7, 7),
            (8, input_dim * 4, 5, 7, 7),
        )
        for shape in shapes:
            yield torch.rand(shape)

    def test_create_simple_module(self):
        """
        Test simple ResNetBasicStem (without pooling layer).
        """
        for input_dim, output_dim in itertools.product((2, 3), (4, 8, 16)):
            model = ResNetBasic3DModule(
                conv=nn.Conv3d(
                    input_dim,
                    output_dim,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                ),
                norm=nn.BatchNorm3d(output_dim),
                activation=nn.ReLU(),
                pool=None,
            )

            # Test forwarding.
            for tensor in TestResNetBasic3DModule._get_inputs(input_dim):
                if tensor.shape[1] != input_dim:
                    with self.assertRaises(RuntimeError):
                        output_tensor = model(tensor)
                    continue
                else:
                    output_tensor = model(tensor)

                input_shape = tensor.shape
                output_shape = output_tensor.shape
                output_shape_gt = (
                    input_shape[0],
                    output_dim,
                    input_shape[2],
                    input_shape[3],
                    input_shape[4],
                )

                self.assertEqual(
                    output_shape,
                    output_shape_gt,
                    f"Output shape {output_shape} is different from expected shape {output_shape_gt}"
                )

    def test_create_complex_module(self):
        """
        Test complex ResNetBasicStem.
        """
        for input_dim, output_dim in itertools.product((2, 3), (4, 8, 16)):
            model = ResNetBasic3DModule(
                conv=nn.Conv3d(
                    input_dim,
                    output_dim,
                    kernel_size=[3, 7, 7],
                    stride=[1, 2, 2],
                    padding=[1, 3, 3],
                    bias=False,
                ),
                norm=nn.BatchNorm3d(output_dim),
                activation=nn.ReLU(),
                pool=nn.MaxPool3d(
                    kernel_size=[1, 3, 3], stride=[1, 2, 2], padding=[0, 1, 1]
                ),
            )

            # Test forwarding.
            for input_tensor in TestResNetBasic3DModule._get_inputs(input_dim):
                if input_tensor.shape[1] != input_dim:
                    with self.assertRaises(Exception):
                        output_tensor = model(input_tensor)
                    continue
                else:
                    output_tensor = model(input_tensor)

                input_shape = input_tensor.shape
                output_shape = output_tensor.shape

                output_shape_gt = (
                    input_shape[0],
                    output_dim,
                    input_shape[2],
                    (((input_shape[3] - 1) // 2 + 1) - 1) // 2 + 1,
                    (((input_shape[4] - 1) // 2 + 1) - 1) // 2 + 1,
                )

                self.assertEqual(
                    output_shape,
                    output_shape_gt,
                    f"Output shape {output_shape} is different from expected shape {output_shape_gt}"
                )

    def test_create_stem_with_callable(self):
        """
        Test builder `create_res_basic_stem` with callable inputs.
        """
        for (pool, activation, norm) in itertools.product(
            (nn.AvgPool3d, nn.MaxPool3d, None),
            (nn.ReLU, nn.Softmax, nn.Sigmoid, None),
            (nn.BatchNorm3d, None),
        ):
            model = create_resnet_basic_3d_module(
                in_channels=3,
                out_channels=64,
                pool=pool,
                activation=activation,
                norm=norm,
            )
            model_gt = ResNetBasic3DModule(
                conv=nn.Conv3d(
                    3,
                    64,
                    kernel_size=[3, 7, 7],
                    stride=[1, 2, 2],
                    padding=[1, 3, 3],
                    bias=False,
                ),
                norm=None if norm is None else norm(64),
                activation=None if activation is None else activation(),
                pool=None
                if pool is None
                else pool(kernel_size=[1, 3, 3], stride=[1, 2, 2], padding=[0, 1, 1]),
            )

            model.load_state_dict(
                model_gt.state_dict(), strict=True
            )  # explicitly use strict mode.

            # Test forwarding.
            for input_tensor in TestResNetBasic3DModule._get_inputs():
                with torch.no_grad():
                    if input_tensor.shape[1] != 3:
                        with self.assertRaises(RuntimeError):
                            output_tensor = model(input_tensor)
                        continue
                    else:
                        output_tensor = model(input_tensor)
                        output_tensor_gt = model_gt(input_tensor)
                self.assertEqual(
                    output_tensor.shape,
                    output_tensor_gt.shape,
                    f"Output shape {output_tensor.shape} is different from expected shape {output_tensor_gt.shape}"
                )
                self.assertTrue(
                    np.allclose(output_tensor.numpy(), output_tensor_gt.numpy())
                )



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import random
import torch
import torch.nn.functional as F

from torch.autograd import Variable
from torch import nn
from towhee.models.loss.focal_loss import FocalLoss


class TestOperator(unittest.TestCase):
    """
    A test for focal loss.
    """

    def setUp(self):
        x = torch.rand(128, 1000, 8, 4) * random.randint(1, 10)
        self.input = Variable(x)
        l = torch.rand(128, 8, 4) * 1000  # 1000 is classes_num
        l = l.long()
        self.target = Variable(l)

    def test_sam(self):
        output0 = FocalLoss(gamma=0)(self.input, self.target)
        output1 = nn.NLLLoss2d()(F.log_softmax(self.input), self.target)
        self.assertGreaterEqual(1, abs(output0.item() - output1.item()))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from typing import Dict
import threading
from queue import Queue

from towhee.engine.operator_runner.runner_base import RunnerStatus
from towhee.engine.operator_runner.flatmap_runner import FlatMapRunner
from towhee.tests.mock_operators.repeat_operator.repeat_operator import RepeatOperator


DATA_QUEUE = Queue()


class StopFrame:
    pass


class MockReader:
    def __init__(self, queue: Queue):
        self._queue = queue

    def read(self):
        data = self._queue.get()
        if not isinstance(data, StopFrame):
            return data
        else:
            raise StopIteration()


class MockWriter:
    def __init__(self):
        self.res = []

    def write(self, data: Dict):
        self.res.append(data)


def run(runner):
    runner.process()


class TestMapRunner(unittest.TestCase):
    """
    MapRunner test
    """

    def test_flatmap_runner(self):
        data_queue = Queue()
        writer = MockWriter()

        # We
        runner = FlatMapRunner('flattest', 0, 'repeat_operator',
                               'mock_operators', {'repeat': 3},
                               MockReader(data_queue), writer)

        runner.set_op(RepeatOperator(3))
        t = threading.Thread(target=run, args=(runner, ))
        t.start()
        self.assertEqual(runner.status, RunnerStatus.RUNNING)
        data_queue.put({'num': 1})
        data_queue.put({'num': 2})
        data_queue.put({'num': 3})
        data_queue.put(None)
        t.join()
        self.assertEqual(runner.status, RunnerStatus.IDLE)

        t = threading.Thread(target=run, args=(runner, ))
        t.start()
        self.assertEqual(runner.status, RunnerStatus.RUNNING)
        data_queue.put({'num': 4})
        data_queue.put({'num': 5})
        data_queue.put(StopFrame())
        runner.join()
        self.assertEqual(runner.status, RunnerStatus.FINISHED)
        self.assertEqual(len(writer.res), 15)

    def test_flatmap_runner_with_error(self):
        data_queue = Queue()
        writer = MockWriter()
        runner = FlatMapRunner('test', 0, 'repeat_operator',
                               'mock_operators', {'repeat': 3},
                               MockReader(data_queue), writer)

        runner.set_op(RepeatOperator(3))
        t = threading.Thread(target=run, args=(runner, ))
        t.start()
        data_queue.put('error_data')
        runner.join()
        self.assertEqual(runner.status, RunnerStatus.FAILED)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import torch

from towhee.models.layers.multi_scale_attention import MultiScaleAttention


class MultiScaleAttentionTest(unittest.TestCase):
    def test_multi_scale_attention(self):
        seq_len = 21
        c_dim = 10
        msa = MultiScaleAttention(c_dim, num_heads=2)
        fake_input = torch.rand(8, seq_len, c_dim)
        input_shape = (2, 2, 5)
        output, output_shape = msa(fake_input, input_shape)
        self.assertTrue(output.shape == fake_input.shape)

        # Test pooling kernel.
        msa = MultiScaleAttention(
            c_dim,
            num_heads=2,
            stride_q=(2, 2, 1),
        )
        output, output_shape = msa(fake_input, input_shape)
        gt_shape_tensor = torch.rand(8, 6, c_dim)
        gt_output_shape = [1, 1, 5]
        self.assertTrue(output.shape == gt_shape_tensor.shape)
        self.assertTrue(output_shape == gt_output_shape)

        # Test pooling kernel with no cls.
        seq_len = 20
        c_dim = 10
        fake_input = torch.rand(8, seq_len, c_dim)
        msa = MultiScaleAttention(
            c_dim, num_heads=2, stride_q=(2, 2, 1), has_cls_embed=False
        )
        output, output_shape = msa(fake_input, input_shape)
        gt_shape_tensor = torch.rand(8, int(seq_len / 2 / 2), c_dim)
        gt_output_shape = [1, 1, 5]
        self.assertTrue(output.shape == gt_shape_tensor.shape)
        self.assertTrue(output_shape == gt_output_shape)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

model_cfgs = {
    'cfg_mnet' : {
        'name': 'mobilenet0.25',
        'min_sizes': [[16, 32], [64, 128], [256, 512]],
        'steps': [8, 16, 32],
        'variance': [0.1, 0.2],
        'clip': False,
        'loc_weight': 2.0,
        'gpu_train': True,
        'batch_size': 32,
        'ngpu': 1,
        'epoch': 250,
        'decay1': 190,
        'decay2': 220,
        'image_size': 640,
        'target_size': 1600,
        'nms_threshold': 0.4,
        'confidence_threshold':0.5,
        'max_size': 2150,
        'mean': [104, 117, 123],
        'std': [1,1,1],
        'pretrain': True,
        'return_layers': {'stage1': 1, 'stage2': 2, 'stage3': 3},
        'in_channel': 32,
        'out_channel': 64
    },

    'cfg_re50' : {
        'name': 'Resnet50',
        'min_sizes': [[16, 32], [64, 128], [256, 512]],
        'steps': [8, 16, 32],
        'variance': [0.1, 0.2],
        'clip': False,
        'loc_weight': 2.0,
        'gpu_train': True,
        'batch_size': 24,
        'ngpu': 4,
        'epoch': 100,
        'decay1': 70,
        'decay2': 90,
        'image_size': 840,
        'target_size': 1600,
        'nms_threshold': 0.4,
        'confidence_threshold':0.5,
        'max_size': 2150,
        'mean': [104, 117, 123],
        'std': [1,1,1],
        'pretrain': True,
        'return_layers': {'layer2': 1, 'layer3': 2, 'layer4': 3},
        'in_channel': 256,
        'out_channel': 256
    }
}

def build_configs(name: str):
    return model_cfgs[name]



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from queue import Queue
from collections import namedtuple

from towhee.dataframe import DataFrame, Variable
from towhee.engine.operator_io.reader import DataFrameReader, BlockMapDataFrameReader, BatchFrameReader
from towhee.engine.operator_io import create_reader

from towhee.tests.test_util.dataframe_test_util import DfWriter, MultiThreadRunner


def read(reader: DataFrameReader, q: Queue):
    while True:
        try:
            item = reader.read()
            if item:
                q.put(item)
                continue
        except StopIteration:
            break


class TestReader(unittest.TestCase):
    """
    Reader test
    """

    def _create_test_df(self, data_size, name = 'test'):
        df = DataFrame(name)
        data = (Variable('int', 1), Variable('str', 'test'), Variable('float', 0.1))
        t = DfWriter(df, data_size, data=data)
        t.set_sealed_when_stop()
        t.start()
        return df

    def test_block_map_reader(self):

        data_size = 100
        df = self._create_test_df(data_size)
        test = {df.name: {'df': df, 'cols': [('v1', 0), ('v2', 2)]}}
        input_order = ['v1', 'v2']

        map_reader = BlockMapDataFrameReader(test, input_order)

        q = Queue()

        runner = MultiThreadRunner(
            target=read, args=(map_reader, q), thread_num=10)

        runner.start()
        runner.join()

        count = 0
        while not q.empty():
            self.assertEqual(q.get(), {'v1': 1, 'v2': 0.1})
            count += 1
        self.assertEqual(count, data_size)

    def test_block_map_reader_multi(self):

        data_size_long = 10
        df_long = self._create_test_df(data_size_long, name = 'long')
        data_size_short = 5
        df_short = self._create_test_df(data_size_short, name = 'short')

        test = {
            df_long.name: {'df': df_long, 'cols': [('v1_long', 0), ('v2_long', 2)]},
            df_short.name: {'df': df_short, 'cols': [('v1_short', 0), ('v2_short', 2)]}
        }

        input_order = ['v1_long', 'v2_long', 'v1_short', 'v2_short']

        map_reader = BlockMapDataFrameReader(test, input_order)

        q = Queue()

        runner = MultiThreadRunner(
            target=read, args=(map_reader, q), thread_num=10)

        runner.start()
        runner.join()

        count = 0
        while not q.empty():
            count += 1
            if count <= 5:
                self.assertEqual(q.get(), {'v1_long': 1, 'v2_long': 0.1, 'v1_short': 1, 'v2_short': 0.1})
            else:
                self.assertEqual(q.get(), {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None})
        self.assertEqual(count, data_size_long)

    def test_batch_reader(self):

        data_size = 100
        df = self._create_test_df(data_size)
        test = {df.name: {'df': df, 'cols': [('v1', 0), ('v2', 2)]}}
        input_order = ['v1', 'v2']
        batch_reader = BatchFrameReader(test, input_order, 3, 2)

        q = Queue()

        runner = MultiThreadRunner(
            target=read, args=(batch_reader, q), thread_num=10)

        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 50)

        num = 0
        d1_map = {'v1': 1, 'v2': 0.1}
        d1 = namedtuple('input', d1_map.keys())(**d1_map)
        while not q.empty():
            if num < 49:
                self.assertEqual(q.get(), [d1] * 3)
            else:
                self.assertEqual(q.get(), [d1] * 2)
            num += 1

    def test_batch_reader_multi(self):

        data_size_long = 100
        df_long = self._create_test_df(data_size_long, name = 'long')
        data_size_short = 50
        df_short = self._create_test_df(data_size_short, name = 'short')

        test = {
            df_long.name: {'df': df_long, 'cols': [('v1_long', 0), ('v2_long', 2)]},
            df_short.name: {'df': df_short, 'cols': [('v1_short', 0), ('v2_short', 2)]}
        }

        input_order = ['v1_long', 'v2_long', 'v1_short', 'v2_short']
        batch_reader = BatchFrameReader(test, input_order, 3, 2)

        q = Queue()

        runner = MultiThreadRunner(
            target=read, args=(batch_reader, q), thread_num=10)

        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 50)

        num = 0

        map1_1 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': 1, 'v2_short': 0.1}
        map1_2 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': 1, 'v2_short': 0.1}
        map1_3 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': 1, 'v2_short': 0.1}
        d1 = [
            namedtuple('input', map1_1.keys())(**map1_1),
            namedtuple('input', map1_2.keys())(**map1_2),
            namedtuple('input', map1_3.keys())(**map1_3)] #24

        map2_1 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': 1, 'v2_short': 0.1}
        map2_2 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': 1, 'v2_short': 0.1}
        map2_3 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None}
        d2 = [
            namedtuple('input', map2_1.keys())(**map2_1),
            namedtuple('input', map2_2.keys())(**map2_2),
            namedtuple('input', map2_3.keys())(**map2_3)] #1

        map3_1 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None}
        map3_2 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None}
        map3_3 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None}
        d3 = [
            namedtuple('input', map3_1.keys())(**map3_1),
            namedtuple('input', map3_2.keys())(**map3_2),
            namedtuple('input', map3_3.keys())(**map3_3)] # 24

        map4_1 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None}
        map4_2 = {'v1_long': 1, 'v2_long': 0.1, 'v1_short': None, 'v2_short': None}
        d4 = [
            namedtuple('input', map4_1.keys())(**map4_1),
            namedtuple('input', map4_2.keys())(**map4_2)] # 1

        while not q.empty():
            if num >=  49:
                self.assertEqual(q.get(), d4)
            elif num >= 25:
                self.assertEqual(q.get(), d3)
            elif num >= 24:
                self.assertEqual(q.get(), d2)
            else:
                self.assertEqual(q.get(), d1)
            num += 1

    def test_close_reader(self):
        data_size = 100
        df = self._create_test_df(data_size)
        test = {df.name: {'df': df, 'cols': [('v1', 0), ('v2', 2)]}}
        input_order = ['v1', 'v2']
        map_reader = create_reader(test, input_order, 'map')

        q = Queue()

        runner = MultiThreadRunner(
            target=read, args=(map_reader, q), thread_num=10)

        runner.start()
        map_reader.close()
        runner.join()

if __name__ == '__main__':
    unittest.main()



import requests
import os
import random
import sys
import getopt
import time
import subprocess
import re
import yaml
from importlib import import_module

from typing import List, Tuple
from tqdm import tqdm
from threading import Thread
from getpass import getpass

from tempfile import TemporaryFile
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError


class Worker(Thread):
    """
    Worker class to realize multi-threads download.

    Args:
        url (`str`):
            The url of the target files.
        local_dir (`str`):
            The local directory to download to.
        file_name (`str`):
            The name of the file (includes extension).
    """

    def __init__(self, url: str, local_dir: str, file_name: str):
        super().__init__()
        self.url = url
        self.local_dir = local_dir
        self.file_name = file_name

    def run(self):
        # Creating the directory tree to the file.
        if not os.path.exists(os.path.dirname(self.local_dir + self.file_name)):
            try:
                os.makedirs(os.path.dirname(self.local_dir + self.file_name))
            except FileExistsError:
                pass
            except OSError as e:
                raise e

        # Get content.
        try:
            r = requests.get(self.url, stream=True)
            if r.status_code == 429:
                time.sleep(3)
                self.run()
            r.raise_for_status()
        except HTTPError as e:
            raise e

        # Create local files.
        file_size = int(r.headers.get('content-length', 0))
        chunk_size = 1024
        progress_bar = tqdm(total=file_size, unit='iB', unit_scale=True, desc=f'Downloading {self.file_name}')
        with open(self.local_dir + self.file_name, 'wb') as local_file:
            for chunk in r.iter_content(chunk_size=chunk_size):
                local_file.write(chunk)
                progress_bar.update(len(chunk))
        progress_bar.close()


def exists(user: str, repo: str) -> bool:
    """
    Check if a repo exists.

    Args:
        user (`str`):
            The author name.
        repo (`str`):
            The repo name.

    Returns:
        (`bool`)
            return `True` if the repository exists, else `False`.
    """
    try:
        url = f'https://hub.towhee.io/api/v1/repos/{user}/{repo}'
        r = requests.get(url)
        return r.status_code == 200
    except HTTPError as e:
        raise e


def create_token(user: str, password: str, token_name: str) -> Tuple[int, str]:
    """
    Create an account verification token.

    This token allows for avoiding HttpBasicAuth for subsequent calls.

    Args:
        user (`str`):
            The account name.
        password (`str`):
            The account password.
        token_name (`str`):
            The name to be given to the token.

    Returns:
        (`Tuple[int, str]`)
            Return the token id and the sha-1.

    Raises:
        (`HTTPError`)
            Raise the error in request.
    """
    url = f'https://hub.towhee.io/api/v1/users/{user}/tokens'
    data = {'name': token_name}
    try:
        r = requests.post(url, data=data, auth=HTTPBasicAuth(user, password))
        r.raise_for_status()
    except HTTPError as e:
        raise e

    res = r.json()
    token_id = str(res['id'])
    token_sha1 = str(res['sha1'])

    return token_id, token_sha1


def delete_token(user: str, password: str, token_id: int) -> None:
    """
    Delete the token with the given name. Useful for cleanup after changes.

    Args:
        user (`str`);
            The account name.
        password (`str`):
            The account password.
        token_id (`int`):
            The token id.
    """
    url = f'https://hub.towhee.io/api/v1/users/{user}/tokens/{token_id}'
    try:
        r = requests.delete(url, auth=HTTPBasicAuth(user, password))
        r.raise_for_status()
    except HTTPError as e:
        raise e


def create_repo(repo: str, token: str, repo_type: str) -> None:
    """
    Create a repo under the account connected to the passed in token.

    Args:
        repo (`str`):
            Name of the repo to create.
        token (`str`):
            Account verification token.
        repo_type (`str`):
            Which category of repo to create, only one can be used, includes
            ('model', 'operator', 'pipeline', 'dataset').

    Raises:
        (`HTTPError`)
            Raise error in request.
    """

    type_dict = {'model': 1, 'operator': 2, 'pipeline': 3, 'dataset': 4}

    # Commented out things in data that are breaking the creation
    data = {
        'auto_init': True,
        'default_branch': 'main',
        'description': 'This is another test repo',
        'name': repo,
        'private': False,
        'template': False,
        'trust_model': 'default',
        'type': type_dict[repo_type]
    }
    url = 'https://hub.towhee.io/api/v1/user/repos'
    try:
        r = requests.post(url, data=data, headers={'Authorization': 'token ' + token})
        r.raise_for_status()
    except HTTPError as e:
        raise e


def delete_repo(user: str, repo: str, token: str) -> None:
    """
    Delete the repo under the user, values correspond to
    https://www.hub.towhee/<user>/<repo>.

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The name of the repo to be deleted.
        token (`str`):
            Account verification token for that user.

    Raises:
        (`HTTPError`)
            Raise error in request.
    """

    url = f'https://hub.towhee.io/api/v1/repos/{user}/{repo}'
    try:
        r = requests.delete(url, headers={'Authorization': 'token ' + token})
        r.raise_for_status()
    except HTTPError as e:
        raise e


def latest_branch_commit(user: str, repo: str, branch: str) -> str:
    """
    Grab the latest commit for a specific branch.

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.
        branch (`str`):
            The branch name.

    Returns:
        (`str`)
            The branch commit hash cut down to 10 characters.

    Raises:
        (`HTTPError`)
            Raise error in request.
    """

    url = f'https://hub.towhee.io/api/v1/repos/{user}/{repo}/commits?limit=1&page=1&sha={branch}'
    try:
        r = requests.get(url, allow_redirects=True)
        r.raise_for_status()
    except HTTPError as e:
        raise e

    res = r.json()

    return res[0]['sha'][:10]


def obtain_lfs_extensions(user: str, repo: str, branch: str) -> List[str]:
    """
    Download the .gitattributes file from the specified repo in order to figure out
    which files are being tracked by git-lfs.

    Lines that deal with git-lfs take on the following format:

    ```
        *.extension   filter=lfs  merge=lfs ...
    ```

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.
        branch (`str`):
            The branch name.

    Returns:
        (`List[str]`)
            The list of file extentions tracked by git-lfs.
    """
    url = f'https://hub.towhee.io/api/v1/repos/{user}/{repo}/raw/.gitattributes?ref={branch}'
    lfs_files = []

    # Using temporary file in order to avoid double download, cleaner to not split up downloads everywhere.
    with TemporaryFile() as temp_file:
        try:
            r = requests.get(url)
            r.raise_for_status()
        except HTTPError:
            return lfs_files

        temp_file.write(r.content)
        temp_file.seek(0)

        for line in temp_file:
            parts = line.split()
            # We only care if lfs filter is present.
            if b'filter=lfs' in parts[1:]:
                # Removing the `*` in `*.ext`, need work if filtering specific files.
                lfs_files.append(parts[0].decode('utf-8')[1:])

    return lfs_files


def get_file_list(user: str, repo: str, commit: str) -> List[str]:
    """
    Get all the files in the current repo at the given commit.

    This is done through forming a git tree recursively and filtering out all the files.

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.
        commit (`str`):
            The commit to base current existing files.

    Returns:
        (`List[str]`)
            The file paths for the repo

    Raises:
        (`HTTPError`)
            Raise error in request.
    """

    url = f'https://hub.towhee.io/api/v1/repos/{user}/{repo}/git/trees/{commit}?recursive=1'
    file_list = []
    try:
        r = requests.get(url)
        r.raise_for_status()
    except HTTPError as e:
        raise e

    res = r.json()
    # Check each object in the tree
    for file in res['tree']:
        # Ignore directories (they have the type 'tree')
        if file['type'] != 'tree':
            file_list.append(file['path'])

    return file_list


def download_files(user: str, repo: str, branch: str, file_list: List[str], lfs_files: List[str], local_dir: str, install_reqs: bool) -> None:
    """
    Download the files from hub. One url is used for git-lfs files and another for the other files.

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.
        branch (`str`):
            The branch name.
        file_list (`List[str]`):
            The hub file paths.
        lfs_files (`List[str]`):
            The file extensions being tracked by git-lfs.
        local_dir (`str`):
            The local directory to download to.
        install_reqs (`bool`):
            Whether to install packages from requirements.txt

    Raises:
        (`HTTPError`)
            Rasie error in request.
        (`OSError`)
            Raise error in writing file.
    """
    threads = []

    # If the trailing forward slash is missing, add it on.
    if local_dir[-1] != '/':
        local_dir += '/'

    # endswith() can check multiple suffixes if they are a tuple.
    lfs_files = tuple(lfs_files)

    for file_name in file_list:
        # Files dealt with lfs have a different url.
        if file_name.endswith(lfs_files):
            url = f'https://hub.towhee.io/{user}/{repo}/media/branch/{branch}/{file_name}'
        else:
            url = f'https://hub.towhee.io/api/v1/repos/{user}/{repo}/raw/{file_name}?ref={branch}'

        threads.append(Worker(url, local_dir, file_name))
        threads[-1].start()

    for thread in threads:
        thread.join()

    if install_reqs:
        requirements = list(filter(lambda x: re.match(r'(.*/)?requirements.txt', x) is not None, file_list))
        for req in requirements:
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', local_dir + req])


def init_file_structure(user: str, repo: str, repo_type: str) -> None:
    """
    Initialized the file structure with template.

    First clone the repo, then download and rename the template repo file.

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.
        repo_type (`str`):
            Which category of repo to create, only one can be used, includes
            ('model', 'operator', 'pipeline', 'dataset').

    Raises:
        (`HTTPError`)
            Raise error in request.
        (`OSError`)
            Raise error in writing file.
    """
    links = 'https://hub.towhee.io/' + user + '/' + repo + '.git'
    subprocess.call(['git', 'clone', links])
    repo_file_name = repo.replace('-', '_')
    if repo_type == 'operator':
        download_repo('towhee', 'operator-template', 'main', os.getcwd() + '/' + repo)
        os.rename(repo + '/operator_template.py', repo + '/' + repo_file_name + '.py')
        os.rename(repo + '/operator_template.yaml', repo + '/' + repo_file_name + '.yaml')
    elif repo_type == 'pipeline':
        download_repo('towhee', 'pipeline-template', 'main', os.getcwd() + '/' + repo)
        os.rename(repo + '/pipeline_template.yaml', repo + '/' + repo_file_name + '.yaml')


def covert_dic(dicts: dict) -> dict:
    """
    Convert all the values in a dictionary to str and replace char, for example:
     <class 'torch.Tensor'>(unknow type) to torch.Tensor(str type).

    Args:
        dicts (`dict`):
            The dictionary to convert.

    Returns:
        (`dict`)
            The converted dictionary.
    """
    for keys in dicts:
        dicts[keys] = str(dicts[keys]).replace('<class ', '').replace('>', '').replace('\'', '')
    return dicts


def generate_repo_yaml(user: str, repo: str) -> None:
    """
    Generate the yaml of Operator, for example:
    name: 'operator-template'
    labels:
      recommended_framework: ''
      class: ''
      others: ''
    operator: 'towhee/operator-template'
    init:
      model_name: str
    call:
      input:
        img_path: str
      output:
        feature_vector: numpy.ndarray

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.

    Raises:
        (`HTTPError`)
            Raise error in request.
        (`OSError`)
            Raise error in writing file.
    """
    sys.path.append(repo)
    repo_file_name = repo.replace('-', '_')
    # get class name in camel case
    components = repo.split('-')
    class_name = ''.join(x.title() for x in components)
    yaml_file = repo + '/' + repo_file_name + '.yaml'
    operator_name = user + '/' + repo
    # import the class from repo
    cls = getattr(import_module(repo_file_name, repo), class_name)

    init_args = cls.__init__.__annotations__
    try:
        del init_args['return']
    except KeyError:
        pass

    call_func = cls.__call__.__annotations__
    try:
        call_output = call_func.pop('return')
        call_output = call_output.__annotations__
    except KeyError:
        pass
    call_input = call_func

    data = {'name': repo,
            'labels': {
                'recommended_framework': '',
                'class': '',
                'others': ''
            },
            'operator': operator_name,
            'init': covert_dic(init_args),
            'call': {
                'input': covert_dic(call_input),
                'output': covert_dic(call_output)
            }}
    with open(yaml_file, 'w', encoding='utf-8') as outfile:
        yaml.dump(data, outfile, default_flow_style=False, sort_keys=False)


def download_repo(user: str, repo: str, branch: str, local_dir: str, install_reqs: bool = True) -> None:
    """
    Performs a download of the selected repo to specified location.

    First checks to see if lfs is tracking files, then finds all the filepaths
    in the repo and lastly downloads them to the location.

    Args:
        user (`str`):
            The account name.
        repo (`str`):
            The repo name.
        branch (`str`):
            The branch name.
        local_dir (`str`):
            The local directory being downloaded to
        install_reqs (`bool`):
            Whether to install packages from requirements.txt

    Raises:
        (`HTTPError`)
            Raise error in request.
        (`OSError`)
            Raise error in writing file.
    """
    if not exists(user, repo):
        raise ValueError(user + '/' + repo + ' repo doesnt exist.')

    lfs_files = obtain_lfs_extensions(user, repo, branch)
    commit = latest_branch_commit(user, repo, branch)
    file_list = get_file_list(user, repo, commit)
    download_files(user, repo, branch, file_list, lfs_files, local_dir, install_reqs)


def main(argv):
    try:
        opts, _ = getopt.getopt(argv[1:], 'u:p:r:t:b:d:', ['create', 'delete', 'download', 'init-directory',
                                                           'generate-yaml', 'user=', 'password=', 'repo=', 'type=',
                                                           'branch=', 'dir='])
    except getopt.GetoptError:
        print(
            'Usage: hub_tool.py -<manipulate type> -u <user> -p ' +
            '<password> -r <repository> -t <repository type> -b <download branch> -d <download directory>'
        )
        sys.exit(2)
    else:
        if argv[0] not in ['create', 'delete', 'download', 'generate-yaml']:
            print('You must specify one kind of manipulation.')
            sys.exit(2)

    user = ''
    password = ''
    repo = ''
    repo_type = ''
    branch = 'main'
    directory = os.getcwd() + '/test_download/'
    # TODO(Filip) figure out how to store the token
    token_name = random.randint(0, 10000)
    manipulation = argv[0]

    for opt, arg in opts:
        if opt in ['-u', '--user']:
            user = arg
        elif opt in ['-p', '--password']:
            password = arg
        elif opt in ['-r', '--repo']:
            repo = arg
        elif opt in ['-t', '--type']:
            repo_type = arg
        elif opt in ['-d', '--dir']:
            directory = arg

    if manipulation in ('create', 'delete'):
        if not user:
            user = input('Please enter your username: ')
        if not password:
            password = getpass('Please enter your password: ')
        if not repo:
            repo = input('Please enter the repo name: ')
        if not repo_type:
            repo_type = input('Please enter the repo type, choose one from "model | operator | pipeline | dataset": ')

        print('Creating token...')
        token_id, token_hash = create_token(user, password, token_name)
        print('token: ', token_hash)

        if manipulation == 'create':
            print('Creating repo...')
            create_repo(repo, token_hash, repo_type)
            print('Generate with template...')
            init_file_structure(user, repo, repo_type)
            print('Done')

        elif manipulation == 'delete':
            print('Deleting repo...')
            delete_repo(user, repo, token_hash)
            print('Done')

        print('Deleting token...')
        # TODO(Filip) right now this doesnt get done if an exception is raised before it
        delete_token(user, password, token_id)
        print('Done')

    elif manipulation == 'download':
        if not user:
            user = input('Please enter the repo author: ')
        if not repo:
            repo = input('Please enter the repo name: ')
        print('Downloading repo...')
        download_repo(user, repo, branch, directory)

    elif manipulation == 'generate-yaml':
        if not user:
            user = input('Please enter your username: ')
        if not repo:
            repo = input('Please enter the operator repo name: ')
        print('Generating the yaml of repo...')
        generate_repo_yaml(user, repo)
        print('Done')


if __name__ == '__main__':
    main(sys.argv[1:])



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import torch
from torchvision.models import resnet34
from torchvision import transforms
from PIL import Image

from towhee.models.embedding.embedding_extractor import EmbeddingExtractor


class EmbeddingExtractorTest(unittest.TestCase):
    def test_embed_extract(self):
        device = torch.device('cpu')
        res = resnet34(pretrained=True)
        res = res.to(device)
        emb = EmbeddingExtractor(res)

        layer_name = 'conv1'
        emb.register(layer_name)

        image = Image.open('cat.jpg')
        transform = transforms.Compose([transforms.Resize((224, 224)), transforms.ToTensor()])
        img = transform(image).unsqueeze(dim=0).to(device)
        out = res(img)
        self.assertTrue(emb.emb_out.embeddings[0].shape == torch.Size([1, 1000]))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import threading
from collections import namedtuple
from typing import Dict, Tuple, Union, List


class ReaderBase(ABC):
    """
    The reader base class.

    The read() could be blocking or non-blocking function, if it's a blocking function,
    the runner may be blocked. When need to stop the graph, we call close to interrupting it.
    """

    @abstractmethod
    def read(self):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError


class DataFrameReader(ReaderBase):
    """
    Read data from input dataframes, unpack and combine data.
    One op_ctx has one dataframe reader.
    """

    def __init__(self, input_readers: List[iter], input_order: List):
        self._iters = input_readers
        self._iters_count = len(self._iters)
        self._input_order = input_order

    @property
    def size(self) -> int:
        return min([x.accessible_size for x in self._iters])

    @property
    def input_order(self) -> List:
        return self._input_order

    @property
    def iters_count(self) -> int:
        return self._iters_count

    @abstractmethod
    def read(self) -> Union[Dict[str, any], List[Dict[str, any]]]:
        pass

    @abstractmethod
    def close(self):
        raise NotImplementedError


class BlockMapReaderWithOriginData(DataFrameReader):
    """
    Return both op's input data and origin data.
    """

    def __init__(self, input_dfs: dict, input_order: List):
        iters = []
        for _, x in input_dfs.items():
            ite = x['df'].map_iter(block=True)
            iters.append([ite, x['cols']])

        super().__init__(iters, input_order)
        self._lock = threading.Lock()
        self._close = False

    def read(self) -> Tuple[Dict[str, any], Tuple]:
        """
        Read data from dataframe.

        Can read data from multi dataframes or a single one. If reading from multiple,
        shorter df's will fill their values with None.
        """
        if self._close:
            raise StopIteration

        ret = {}
        origin_data = ()
        count_stop = self.iters_count

        with self._lock:
            for x in self._iters:
                try:
                    data = next(x[0])
                except StopIteration:
                    # If stop iteration raised, fill data with None
                    if self._close:
                        raise StopIteration  #pylint: disable=raise-missing-from
                    for (key, index) in x[1]:
                        ret[key] = None
                    origin_data += (None,)
                    count_stop -= 1
                    continue

                if self._close:
                    raise StopIteration
                if data is not None:
                    for (key, index) in x[1]:
                        ret[key] = data[index].value
                    origin_data += data

            if count_stop == 0:
                raise StopIteration
            # TODO: origin_data ordering with filter
            return ret, origin_data

    def close(self):
        self._close = True
        for x in self._iters:
            x[0].notify()


class BlockMapDataFrameReader(BlockMapReaderWithOriginData):
    """
    Map dataframe reader.
    """

    def read(self) -> Dict[str, any]:
        output, _ = super().read()
        return output


class BatchFrameReader(DataFrameReader):
    """
    Batch reader.

    Reads batches of data from one or multiple dataframes. Missing values will be
    replaced with none.
    """

    def __init__(self, input_dfs: dict, input_order: List, batch_size: int, step: int):
        assert batch_size >= 1 and step >= 1
        iters = []
        for _, x in input_dfs.items():
            ite = x['df'].batch_iter(batch_size, step, block=True)
            iters.append([ite, x['cols']])

        super().__init__(iters, input_order)
        self._lock = threading.Lock()
        self._close = False

    def read(self) -> List[Dict[str, any]]:
        if self._close:
            raise StopIteration

        ret = {}
        count_stop = self.iters_count

        with self._lock:
            for x in self._iters:
                try:
                    data = next(x[0])
                except StopIteration:
                    count_stop -= 1
                    if self._close:
                        raise StopIteration #pylint: disable=raise-missing-from
                    continue

                if self._close:
                    raise StopIteration
                if data is not None:
                    for i, row in enumerate(data):
                        if i not in ret:
                            ret[i] = {}
                        for (key, index) in x[1]:
                            ret[i][key] = row[index].value

            if count_stop == 0:
                raise StopIteration

            # Fill missing values with Nones
            # TODO: Figure out faster/better logic for this
            for key, val in ret.items():
                for x in self.input_order:
                    if x not in val:
                        val[x] = None

            res = []
            for i in range(len(ret)):
                res.append(namedtuple('input', ret[i].keys())(**ret[i]))
            return res

    def close(self):
        self._close = True
        for x in self._iters:
            x[0].notify()



# coding : UTF-8
from operator import methodcaller
from test_image_embedding import *
from test_pipeline import *
from test_image_object_embedding import *

def pipeline_register():

    pipeline_names = ["image-embedding", "towhee/video_embedding_resnet50",
                      "towhee/img_object_embedding"]

    return pipeline_names

def pipeline_runner():

    invalid_pipeline_obj = TestPipelineInvalid()
    for func in dir(TestPipelineInvalid):
        if not func.startswith("__"):
            res = methodcaller(func)(invalid_pipeline_obj)
            if res == None:
                print("%s PASS" % func)
            else:
                print("%s FAIL" % func)

    pipeline_names = pipeline_register()
    for pipeline_name in pipeline_names:
        valid_pipeline_obj = TestPipelineValid()
        for func in dir(TestPipelineValid):
            if not func.startswith("__"):
                res = methodcaller(func, pipeline_name)(valid_pipeline_obj)
                if res == None:
                    print("%s:%s PASS" % (func, pipeline_name))
                else:
                    print("%s:%s FAIL" % (func, pipeline_name))
    return True

def image_class_pipeline_register():

    pipeline_names = ["image-embedding"]

    return pipeline_names

def image_class_pipeline_runner():

    pipeline_names = image_class_pipeline_register()
    for pipeline_name in pipeline_names:
        invalid_embedding_obj = TestImageEmbeddingInvalid()
        for func in dir(TestImageEmbeddingInvalid):
            if not func.startswith("__"):
                res = methodcaller(func, pipeline_name)(invalid_embedding_obj)
                if res == 1:
                    print("%s:%s PASS" % (func, pipeline_name))
                else:
                    print("%s:%s FAIL" % (func, pipeline_name))
        valid_embedding_obj = TestImageEmbeddingValid()
        for func in dir(TestImageEmbeddingValid):
            if not func.startswith("__"):
                res = methodcaller(func, pipeline_name)(valid_embedding_obj)
                if res == 1:
                    print("%s:%s PASS" % (func, pipeline_name))
                else:
                    print("%s:%s FAIL" % (func, pipeline_name))

    return True

def image_obj_class_pipeline_register():

    pipeline_names = ["towhee/img_object_embedding"]

    return pipeline_names

def image_obj_class_pipeline_runner():

    pipeline_names = image_obj_class_pipeline_register()
    for pipeline_name in pipeline_names:
        invalid_embedding_obj = TestImageObjEmbeddingInvalid()
        for func in dir(TestImageObjEmbeddingInvalid):
            if not func.startswith("__"):
                res = methodcaller(func, pipeline_name)(invalid_embedding_obj)
                if res == 1:
                    print("%s:%s PASS" % (func, pipeline_name))
                else:
                    print("%s:%s FAIL" % (func, pipeline_name))
        valid_embedding_obj = TestImageObjEmbeddingValid()
        for func in dir(TestImageObjEmbeddingValid):
            if not func.startswith("__"):
                res = methodcaller(func, pipeline_name)(valid_embedding_obj)
                if res == 1:
                    print("%s:%s PASS" % (func, pipeline_name))
                else:
                    print("%s:%s FAIL" % (func, pipeline_name))

    return True

def test_caller():

    pipeline_runner()
    image_class_pipeline_runner()
    image_obj_class_pipeline_runner()

    return True


if __name__ == '__main__':

    test_caller()






